@@ -6,8 +6,7 @@
 #include <stdio.h>
 #include <string.h>
 
-#include <checkers.h>
-
+#include "checkers.h"
 #include "memory.h"
 #include "util.h"
 #include "debug.h"
@@ -19,24 +18,33 @@
 #include "config.h"
 #include "blacklist.h"
 #include "defaults.h"
+#include "prio.h"
+
+static int
+hwe_strmatch (struct hwentry *hwe1, struct hwentry *hwe2)
+{
+	if (hwe1->vendor && hwe2->vendor && strcmp(hwe1->vendor, hwe2->vendor))
+		return 1;
+
+	if (hwe1->product && hwe2->product && strcmp(hwe1->product, hwe2->product))
+		return 1;
+
+	if (hwe1->revision && hwe2->revision && strcmp(hwe1->revision, hwe2->revision))
+		return 1;
+
+	return 0;
+}
 
 static struct hwentry *
-find_hwe_strmatch (vector hwtable, char * vendor, char * product, char * revision)
+find_hwe_strmatch (vector hwtable, struct hwentry *hwe)
 {
 	int i;
-	struct hwentry *hwe, *ret = NULL;
+	struct hwentry *tmp, *ret = NULL;
 
-	vector_foreach_slot (hwtable, hwe, i) {
-		if (hwe->vendor && vendor && strcmp(hwe->vendor, vendor))
+	vector_foreach_slot (hwtable, tmp, i) {
+		if (hwe_strmatch(tmp, hwe))
 			continue;
-
-		if (hwe->product && product && strcmp(hwe->product, product))
-			continue;
-
-		if (hwe->revision && revision && strcmp(hwe->revision, revision))
-			continue;
-
-		ret = hwe;
+		ret = tmp;
 		break;
 	}
 	return ret;
@@ -129,21 +137,27 @@ free_hwe (struct hwentry * hwe)
 	if (hwe->revision)
 		FREE(hwe->revision);
 
-	if (hwe->selector)
-		FREE(hwe->selector);
-
 	if (hwe->getuid)
 		FREE(hwe->getuid);
 
-	if (hwe->getprio)
-		FREE(hwe->getprio);
-
 	if (hwe->features)
 		FREE(hwe->features);
 
 	if (hwe->hwhandler)
 		FREE(hwe->hwhandler);
 
+	if (hwe->selector)
+		FREE(hwe->selector);
+
+	if (hwe->checker_name)
+		FREE(hwe->checker_name);
+
+	if (hwe->prio_name)
+		FREE(hwe->prio_name);
+
+	if (hwe->prio_args)
+		FREE(hwe->prio_args);
+
 	if (hwe->bl_product)
 		FREE(hwe->bl_product);
 
@@ -210,7 +224,7 @@ alloc_mpe (void)
 	return mpe;
 }
 
-static struct hwentry *
+struct hwentry *
 alloc_hwe (void)
 {
 	struct hwentry * hwe = (struct hwentry *)
@@ -242,47 +256,88 @@ set_param_str(char * str)
 	return dst;
 }
 
+#define merge_str(s) \
+	if (hwe2->s) { \
+		if (hwe1->s) \
+			FREE(hwe1->s); \
+		if (!(hwe1->s = set_param_str(hwe2->s))) \
+			return 1; \
+	}
+
+#define merge_num(s) \
+	if (hwe2->s) \
+		hwe1->s = hwe2->s
+
+
+static int
+merge_hwe (struct hwentry * hwe1, struct hwentry * hwe2)
+{
+	merge_str(vendor);
+	merge_str(product);
+	merge_str(revision);
+	merge_str(getuid);
+	merge_str(features);
+	merge_str(hwhandler);
+	merge_str(selector);
+	merge_str(checker_name);
+	merge_str(prio_name);
+	merge_str(prio_args);
+	merge_str(bl_product);
+	merge_num(pgpolicy);
+	merge_num(pgfailback);
+	merge_num(rr_weight);
+	merge_num(no_path_retry);
+	merge_num(minio);
+
+	return 0;
+}
+
 int
 store_hwe (vector hwtable, struct hwentry * dhwe)
 {
 	struct hwentry * hwe;
 
-	if (find_hwe_strmatch(hwtable, dhwe->vendor, dhwe->product, dhwe->revision))
+	if (find_hwe_strmatch(hwtable, dhwe))
 		return 0;
-	
+
 	if (!(hwe = alloc_hwe()))
 		return 1;
 
 	if (!dhwe->vendor || !(hwe->vendor = set_param_str(dhwe->vendor)))
 		goto out;
-	
+
 	if (!dhwe->product || !(hwe->product = set_param_str(dhwe->product)))
 		goto out;
-	
+
 	if (dhwe->revision && !(hwe->revision = set_param_str(dhwe->revision)))
 		goto out;
-	
+
 	if (dhwe->getuid && !(hwe->getuid = set_param_str(dhwe->getuid)))
 		goto out;
 
-	if (dhwe->getprio && !(hwe->getprio = set_param_str(dhwe->getprio)))
-		goto out;
-				
 	if (dhwe->features && !(hwe->features = set_param_str(dhwe->features)))
 		goto out;
-	
+
 	if (dhwe->hwhandler && !(hwe->hwhandler = set_param_str(dhwe->hwhandler)))
 		goto out;
 
 	if (dhwe->selector && !(hwe->selector = set_param_str(dhwe->selector)))
 		goto out;
-				
+
+	if (dhwe->checker_name && !(hwe->checker_name = set_param_str(dhwe->checker_name)))
+		goto out;
+
+	if (dhwe->prio_name && !(hwe->prio_name = set_param_str(dhwe->prio_name)))
+		goto out;
+
+	if (dhwe->prio_args && !(hwe->prio_args = set_param_str(dhwe->prio_args)))
+		goto out;
+
 	hwe->pgpolicy = dhwe->pgpolicy;
 	hwe->pgfailback = dhwe->pgfailback;
 	hwe->rr_weight = dhwe->rr_weight;
 	hwe->no_path_retry = dhwe->no_path_retry;
 	hwe->minio = dhwe->minio;
-	hwe->checker = dhwe->checker;
 
 	if (dhwe->bl_product && !(hwe->bl_product = set_param_str(dhwe->bl_product)))
 		goto out;
@@ -297,6 +352,27 @@ out:
 	return 1;
 }
 
+static int
+factorize_hwtable (vector hw)
+{
+	struct hwentry *hwe1, *hwe2;
+	int i, j;
+
+	vector_foreach_slot(hw, hwe1, i) {
+		j = i+1;
+		vector_foreach_slot_after(hw, hwe2, j) {
+			if (hwe_strmatch(hwe1, hwe2))
+				continue;
+			/* dup */
+			merge_hwe(hwe1, hwe2);
+			free_hwe(hwe2);
+			vector_del_slot(hw, j);
+			j--;
+		}
+	}
+	return 0;
+}
+
 struct config *
 alloc_config (void)
 {
@@ -315,21 +391,30 @@ free_config (struct config * conf)
 	if (conf->udev_dir)
 		FREE(conf->udev_dir);
 
+	if (conf->multipath_dir)
+		FREE(conf->multipath_dir);
+
 	if (conf->selector)
 		FREE(conf->selector);
 
 	if (conf->getuid)
 		FREE(conf->getuid);
 
-	if (conf->getprio)
-		FREE(conf->getprio);
-
 	if (conf->features)
 		FREE(conf->features);
 
 	if (conf->hwhandler)
 		FREE(conf->hwhandler);
 
+	if (conf->prio_name)
+		FREE(conf->prio_name);
+
+	if (conf->prio_args)
+		FREE(conf->prio_args);
+
+	if (conf->checker_name)
+		FREE(conf->checker_name);
+
 	free_blacklist(conf->blist_devnode);
 	free_blacklist(conf->blist_wwid);
 	free_blacklist_device(conf->blist_device);
@@ -357,11 +442,27 @@ load_config (char * file)
 	 * internal defaults
 	 */
 	if (!conf->verbosity)
-		conf->verbosity = 2;
+		conf->verbosity = DEFAULT_VERBOSITY;
 
 	conf->dev_type = DEV_NONE;
 	conf->minio = 1000;
+	conf->max_fds = 0;
 	conf->bindings_file = DEFAULT_BINDINGS_FILE;
+	conf->multipath_dir = set_default(DEFAULT_MULTIPATHDIR);
+	conf->flush_on_last_del = 0;
+	conf->attribute_flags = 0;
+
+	/*
+	 * preload default hwtable
+	 */
+	if (conf->hwtable == NULL) {
+		conf->hwtable = vector_alloc();
+
+		if (!conf->hwtable)
+			goto out;
+	}
+	if (setup_default_hwtable(conf->hwtable))
+		goto out;
 
 	/*
 	 * read the config file
@@ -373,35 +474,31 @@ load_config (char * file)
 			goto out;
 		}
 	}
-	
+
+	/*
+	 * remove duplica in hwtable. config file takes precedence
+	 * over build-in hwtable
+	 */
+	factorize_hwtable(conf->hwtable);
+
 	/*
 	 * fill the voids left in the config file
 	 */
-	if (conf->hwtable == NULL) {
-		conf->hwtable = vector_alloc();
-		
-		if (!conf->hwtable)
-			goto out;
-		
-	}
-	if (setup_default_hwtable(conf->hwtable))
-		goto out;
-
 	if (conf->blist_devnode == NULL) {
 		conf->blist_devnode = vector_alloc();
-		
+
 		if (!conf->blist_devnode)
 			goto out;
 	}
 	if (conf->blist_wwid == NULL) {
 		conf->blist_wwid = vector_alloc();
-		
+
 		if (!conf->blist_wwid)
 			goto out;
 	}
 	if (conf->blist_device == NULL) {
 		conf->blist_device = vector_alloc();
-		
+
 		if (!conf->blist_device)
 			goto out;
 	}
@@ -409,54 +506,36 @@ load_config (char * file)
 		goto out;
 
 	if (conf->elist_devnode == NULL) {
-                conf->elist_devnode = vector_alloc();
+		conf->elist_devnode = vector_alloc();
 
-                if (!conf->elist_devnode)
+		if (!conf->elist_devnode)
 			goto out;
 	}
 	if (conf->elist_wwid == NULL) {
 		conf->elist_wwid = vector_alloc();
 
-                if (!conf->elist_wwid)
+		if (!conf->elist_wwid)
 			goto out;
 	}
 
 	if (conf->elist_device == NULL) {
 		conf->elist_device = vector_alloc();
-		
+
 		if (!conf->elist_device)
 			goto out;
 	}
 
 	if (conf->mptable == NULL) {
 		conf->mptable = vector_alloc();
-
 		if (!conf->mptable)
 			goto out;
 	}
-	if (conf->selector == NULL)
-		conf->selector = set_default(DEFAULT_SELECTOR);
-
 	if (conf->udev_dir == NULL)
 		conf->udev_dir = set_default(DEFAULT_UDEVDIR);
 
-	if (conf->getuid == NULL)
-		conf->getuid = set_default(DEFAULT_GETUID);
-
-	if (conf->features == NULL)
-		conf->features = set_default(DEFAULT_FEATURES);
-
-	if (conf->hwhandler == NULL)
-		conf->hwhandler = set_default(DEFAULT_HWHANDLER);
-
-	if (!conf->selector  || !conf->udev_dir         ||
-	    !conf->getuid    || !conf->features ||
-	    !conf->hwhandler)
+	if (!conf->udev_dir || !conf->multipath_dir)
 		goto out;
 
-	if (!conf->checker)
-		conf->checker = checker_lookup(DEFAULT_CHECKER);
-
 	return 0;
 out:
 	free_config(conf);
