@@ -1,6 +1,9 @@
 #ifndef _CONFIG_H
 #define _CONFIG_H
 
+#include <sys/types.h>
+#include <stdint.h>
+
 #define ORIGIN_DEFAULT 0
 #define ORIGIN_CONFIG  1
 
@@ -16,11 +19,12 @@ struct hwentry {
 	char * product;
 	char * revision;
 	char * getuid;
-	char * getprio;
 	char * features;
 	char * hwhandler;
 	char * selector;
 	char * checker_name;
+	char * prio_name;
+	char * prio_args;
 
 	int pgpolicy;
 	int pgfailback;
@@ -28,7 +32,9 @@ struct hwentry {
 	int no_path_retry;
 	int minio;
 	int pg_timeout;
-	struct checker * checker;
+	int flush_on_last_del;
+	int fast_io_fail;
+	unsigned int dev_loss;
 	char * bl_product;
 };
 
@@ -44,6 +50,11 @@ struct mpentry {
 	int no_path_retry;
 	int minio;
 	int pg_timeout;
+	int flush_on_last_del;
+	int attribute_flags;
+	uid_t uid;
+	gid_t gid;
+	mode_t mode;
 };
 
 struct config {
@@ -53,7 +64,6 @@ struct config {
 	int pgpolicy_flag;
 	int with_sysfs;
 	int pgpolicy;
-	struct checker * checker;
 	enum devtypes dev_type;
 	int minio;
 	int checkint;
@@ -64,16 +74,31 @@ struct config {
 	int no_path_retry;
 	int user_friendly_names;
 	int pg_timeout;
+	int max_fds;
+	int force_reload;
+	int queue_without_daemon;
+	int daemon;
+	int flush_on_last_del;
+	int attribute_flags;
+	int fast_io_fail;
+	unsigned int dev_loss;
+	uid_t uid;
+	gid_t gid;
+	mode_t mode;
+	uint32_t cookie;
 
 	char * dev;
 	char * sysfs_dir;
 	char * udev_dir;
+	char * multipath_dir;
 	char * selector;
 	char * getuid;
-	char * getprio;
 	char * features;
 	char * hwhandler;
 	char * bindings_file;
+	char * prio_name;
+	char * prio_args;
+	char * checker_name;
 
 	vector keywords;
 	vector mptable;
@@ -93,6 +118,7 @@ struct hwentry * find_hwe (vector hwtable, char * vendor, char * product, char *
 struct mpentry * find_mpe (char * wwid);
 char * get_mpe_wwid (char * alias);
 
+struct hwentry * alloc_hwe (void);
 struct mpentry * alloc_mpe (void);
 
 void free_hwe (struct hwentry * hwe);
