@@ -14,8 +14,7 @@
 #include <errno.h>
 #include <libdevmapper.h>
 
-#include <checkers.h>
-
+#include "checkers.h"
 #include "vector.h"
 #include "memory.h"
 #include "devmapper.h"
@@ -34,6 +33,8 @@
 #include "pgpolicies.h"
 #include "dict.h"
 #include "alias.h"
+#include "prio.h"
+#include "util.h"
 
 extern int
 setup_map (struct multipath * mpp)
@@ -49,6 +50,11 @@ setup_map (struct multipath * mpp)
 		return 1;
 	}
 
+	/*
+	 * free features, selector, and hwhandler properties if they are being reused
+	 */
+	free_multipath_attributes(mpp);
+
 	/*
 	 * properties selectors
 	 */
@@ -61,7 +67,13 @@ setup_map (struct multipath * mpp)
 	select_minio(mpp);
 	select_no_path_retry(mpp);
 	select_pg_timeout(mpp);
+	select_mode(mpp);
+	select_uid(mpp);
+	select_gid(mpp);
+	select_fast_io_fail(mpp);
+	select_dev_loss(mpp);
 
+	sysfs_set_scsi_tmo(mpp);
 	/*
 	 * assign paths to path groups -- start with no groups and all paths
 	 * in mpp->paths
@@ -113,6 +125,9 @@ pgcmp (struct multipath * mpp, struct multipath * cmpp)
 	struct pathgroup * cpgp;
 	int r = 0;
 
+	if (!mpp)
+		return 0;
+
 	vector_foreach_slot (mpp->pg, pgp, i) {
 		compute_pgid(pgp);
 
@@ -130,7 +145,7 @@ pgcmp (struct multipath * mpp, struct multipath * cmpp)
 }
 
 static void
-select_action (struct multipath * mpp, vector curmp)
+select_action (struct multipath * mpp, vector curmp, int force_reload)
 {
 	struct multipath * cmpp;
 
@@ -139,7 +154,7 @@ select_action (struct multipath * mpp, vector curmp)
 	if (!cmpp) {
 		cmpp = find_mp_by_wwid(curmp, mpp->wwid);
 
-		if (cmpp && !conf->dry_run) {
+		if (cmpp) {
 			condlog(2, "%s: rename %s to %s", mpp->wwid,
 				cmpp->alias, mpp->alias);
 			strncpy(mpp->alias_old, cmpp->alias, WWID_SIZE);
@@ -154,7 +169,7 @@ select_action (struct multipath * mpp, vector curmp)
 
 	if (!find_mp_by_wwid(curmp, mpp->wwid)) {
 		condlog(2, "%s: remove (wwid changed)", cmpp->alias);
-		dm_flush_map(mpp->alias, DEFAULT_TARGET);
+		dm_flush_map(mpp->alias);
 		strncat(cmpp->wwid, mpp->wwid, WWID_SIZE);
 		drop_multipath(curmp, cmpp->wwid, KEEP_PATHS);
 		mpp->action = ACT_CREATE;
@@ -162,13 +177,19 @@ select_action (struct multipath * mpp, vector curmp)
 			mpp->alias);
 		return;
 	}
-		
+
 	if (pathcount(mpp, PATH_UP) == 0) {
 		mpp->action = ACT_NOTHING;
 		condlog(3, "%s: set ACT_NOTHING (no usable path)",
 			mpp->alias);
 		return;
 	}
+	if (force_reload) {
+		mpp->action = ACT_RELOAD;
+		condlog(3, "%s: set ACT_RELOAD (forced by user)",
+			mpp->alias);
+		return;
+	}
 	if (cmpp->size != mpp->size) {
 		mpp->action = ACT_RELOAD;
 		condlog(3, "%s: set ACT_RELOAD (size change)",
@@ -183,14 +204,14 @@ select_action (struct multipath * mpp, vector curmp)
 			mpp->alias);
 		return;
 	}
-	if (strncmp(cmpp->hwhandler, mpp->hwhandler,
+	if (!cmpp->selector || strncmp(cmpp->hwhandler, mpp->hwhandler,
 		    strlen(mpp->hwhandler))) {
 		mpp->action = ACT_RELOAD;
 		condlog(3, "%s: set ACT_RELOAD (hwhandler change)",
 			mpp->alias);
 		return;
 	}
-	if (strncmp(cmpp->selector, mpp->selector,
+	if (!cmpp->selector || strncmp(cmpp->selector, mpp->selector,
 		    strlen(mpp->selector))) {
 		mpp->action = ACT_RELOAD;
 		condlog(3, "%s: set ACT_RELOAD (selector change)",
@@ -203,7 +224,7 @@ select_action (struct multipath * mpp, vector curmp)
 			mpp->alias, cmpp->minio, mpp->minio);
 		return;
 	}
-	if (VECTOR_SIZE(cmpp->pg) != VECTOR_SIZE(mpp->pg)) {
+	if (!cmpp->pg || VECTOR_SIZE(cmpp->pg) != VECTOR_SIZE(mpp->pg)) {
 		mpp->action = ACT_RELOAD;
 		condlog(3, "%s: set ACT_RELOAD (path group number change)",
 			mpp->alias);
@@ -266,7 +287,7 @@ lock_multipath (struct multipath * mpp, int lock)
 
 	if (!mpp || !mpp->pg)
 		return 0;
-	
+
 	vector_foreach_slot (mpp->pg, pgp, i) {
 		if (!pgp->paths)
 			continue;
@@ -327,31 +348,32 @@ domap (struct multipath * mpp)
 
 		if (dm_map_present(mpp->alias)) {
 			condlog(3, "%s: map already present", mpp->alias);
+			lock_multipath(mpp, 0);
 			break;
 		}
 
-		r = dm_addmap(DM_DEVICE_CREATE, mpp->alias, DEFAULT_TARGET,
-			      mpp->params, mpp->size, mpp->wwid);
+		r = dm_addmap_create(mpp);
 
-		/*
-		 * DM_DEVICE_CREATE is actually DM_DEV_CREATE plus
-		 * DM_TABLE_LOAD. Failing the second part leaves an
-		 * empty map. Clean it up.
-		 */
-		if (!r && dm_map_present(mpp->alias)) {
-			condlog(3, "%s: failed to load map "
-				   "(a path might be in use)",
-				   mpp->alias);
-			dm_flush_map(mpp->alias, DEFAULT_TARGET);
-		}
+		if (!r)
+			 r = dm_addmap_create_ro(mpp);
 
 		lock_multipath(mpp, 0);
 		break;
 
 	case ACT_RELOAD:
-		r = (dm_addmap(DM_DEVICE_RELOAD, mpp->alias, DEFAULT_TARGET,
-			      mpp->params, mpp->size, NULL) &&
-		     dm_simplecmd(DM_DEVICE_RESUME, mpp->alias));
+		r = dm_addmap_reload(mpp);
+		if (!r)
+			r = dm_addmap_reload_ro(mpp);
+		if (r)
+			r = dm_simplecmd_noflush(DM_DEVICE_RESUME, mpp->alias);
+		break;
+
+ 	case ACT_RESIZE:
+  		r = dm_addmap_reload(mpp);
+  		if (!r)
+  			r = dm_addmap_reload_ro(mpp);
+  		if (r)
+  			r = dm_simplecmd_flush(DM_DEVICE_RESUME, mpp->alias, 1);
 		break;
 
 	case ACT_RENAME:
@@ -364,18 +386,27 @@ domap (struct multipath * mpp)
 
 	if (r) {
 		/*
-		 * DM_DEVICE_CREATE, DM_DEIVCE_RENAME, or DM_DEVICE_RELOAD
+		 * DM_DEVICE_CREATE, DM_DEVICE_RENAME, or DM_DEVICE_RELOAD
 		 * succeeded
 		 */
-#ifndef DAEMON
-		dm_switchgroup(mpp->alias, mpp->bestpg);
-		if (mpp->action != ACT_NOTHING)
-			print_multipath_topology(mpp, conf->verbosity);
-#else
-		mpp->stat_map_loads++;
-		condlog(2, "%s: load table [0 %llu %s %s]", mpp->alias,
-                        mpp->size, DEFAULT_TARGET, mpp->params);
-#endif
+		if (!conf->daemon) {
+			/* multipath client mode */
+			dm_switchgroup(mpp->alias, mpp->bestpg);
+			if (mpp->action != ACT_NOTHING)
+				print_multipath_topology(mpp, conf->verbosity);
+		} else  {
+			/* multipath daemon mode */
+			mpp->stat_map_loads++;
+			condlog(2, "%s: load table [0 %llu %s %s]", mpp->alias,
+				mpp->size, TGT_MPATH, mpp->params);
+			/*
+			 * Required action is over, reset for the stateful daemon.
+			 * But don't do it for creation as we use in the caller the
+			 * mpp->action to figure out whether to start the watievent checker.
+			 */
+			if (mpp->action != ACT_CREATE)
+				mpp->action = ACT_NOTHING;
+		}
 		return DOMAP_OK;
 	}
 	return DOMAP_FAIL;
@@ -399,12 +430,12 @@ deadmap (struct multipath * mpp)
 			if (strlen(pp->dev))
 				return 0; /* alive */
 	}
-	
+
 	return 1; /* dead */
 }
 
 extern int
-coalesce_paths (struct vectors * vecs, vector newmp, char * refwwid)
+coalesce_paths (struct vectors * vecs, vector newmp, char * refwwid, int force_reload)
 {
 	int r = 1;
 	int k, i;
@@ -417,6 +448,11 @@ coalesce_paths (struct vectors * vecs, vector newmp, char * refwwid)
 
 	memset(empty_buff, 0, WWID_SIZE);
 
+	if (force_reload) {
+		vector_foreach_slot (pathvec, pp1, k) {
+			pp1->mpp = NULL;
+		}
+	}
 	vector_foreach_slot (pathvec, pp1, k) {
 		/* skip this path for some reason */
 
@@ -440,7 +476,8 @@ coalesce_paths (struct vectors * vecs, vector newmp, char * refwwid)
 		/*
 		 * at this point, we know we really got a new mp
 		 */
-		if ((mpp = add_map_with_path(vecs, pp1, 0)) == NULL)
+		mpp = add_map_with_path(vecs, pp1, 0);
+		if (!mpp)
 			return 1;
 
 		if (pp1->priority == PRIO_UNDEF)
@@ -448,16 +485,16 @@ coalesce_paths (struct vectors * vecs, vector newmp, char * refwwid)
 
 		if (!mpp->paths) {
 			condlog(0, "%s: skip coalesce (no paths)", mpp->alias);
-			remove_map(mpp, vecs, NULL, 0);
+			remove_map(mpp, vecs, 0);
 			continue;
 		}
-		
+
 		for (i = k + 1; i < VECTOR_SIZE(pathvec); i++) {
 			pp2 = VECTOR_SLOT(pathvec, i);
 
 			if (strcmp(pp1->wwid, pp2->wwid))
 				continue;
-			
+
 			if (!pp2->size)
 				continue;
 
@@ -476,12 +513,12 @@ coalesce_paths (struct vectors * vecs, vector newmp, char * refwwid)
 		verify_paths(mpp, vecs, NULL);
 		
 		if (setup_map(mpp)) {
-			remove_map(mpp, vecs, NULL, 0);
+			remove_map(mpp, vecs, 0);
 			continue;
 		}
 
 		if (mpp->action == ACT_UNDEF)
-			select_action(mpp, curmp);
+			select_action(mpp, curmp, force_reload);
 
 		r = domap(mpp);
 
@@ -490,7 +527,7 @@ coalesce_paths (struct vectors * vecs, vector newmp, char * refwwid)
 				   "for create/reload map",
 				mpp->alias, r);
 			if (r == DOMAP_FAIL) {
-				remove_map(mpp, vecs, NULL, 0);
+				remove_map(mpp, vecs, 0);
 				continue;
 			} else /* if (r == DOMAP_RETRY) */
 				return r;
@@ -518,7 +555,7 @@ coalesce_paths (struct vectors * vecs, vector newmp, char * refwwid)
 				vector_set_slot(newmp, mpp);
 			}
 			else
-				remove_map(mpp, vecs, NULL, 0);
+				remove_map(mpp, vecs, 0);
 		}
 	}
 	/*
@@ -538,9 +575,9 @@ coalesce_paths (struct vectors * vecs, vector newmp, char * refwwid)
 			if ((j = find_slot(newmp, (void *)mpp)) != -1)
 				vector_del_slot(newmp, j);
 
-			remove_map(mpp, vecs, NULL, 0);
+			remove_map(mpp, vecs, 0);
 
-			if (dm_flush_map(mpp->alias, DEFAULT_TARGET))
+			if (dm_flush_map(mpp->alias))
 				condlog(2, "%s: remove failed (dead)",
 					mpp->alias);
 			else
@@ -561,7 +598,7 @@ get_refwwid (char * dev, enum devtypes dev_type, vector pathvec)
 		return NULL;
 
 	if (dev_type == DEV_DEVNODE) {
-		basename(dev, buff);
+		basenamecpy(dev, buff);
 		pp = find_path_by_dev(pathvec, buff);
 		
 		if (!pp) {
