@@ -11,8 +11,7 @@
 #include <dirent.h>
 #include <errno.h>
 
-#include <checkers.h>
-
+#include "checkers.h"
 #include "vector.h"
 #include "memory.h"
 #include "util.h"
@@ -25,6 +24,7 @@
 #include "sg_include.h"
 #include "sysfs.h"
 #include "discovery.h"
+#include "prio.h"
 
 struct path *
 store_pathinfo (vector pathvec, vector hwtable, char * devname, int flag)
@@ -70,7 +70,7 @@ path_discover (vector pathvec, struct config * conf, char * devname, int flag)
 		condlog(0, "path too small");
 		return 1;
 	}
-			
+
 	if (strncmp(devname,"cciss",5) && !filepresent(path)) {
 		condlog(4, "path %s not present", path);
 		return 0;
@@ -133,9 +133,9 @@ sysfs_get_##fname (struct sysfs_device * dev, char * buff, size_t len) \
 	attr = sysfs_attr_get_value(dev->devpath, #fname); \
 	if (!attr) \
 		return 1; \
-\
 	if (strlcpy(buff, attr, len) != strlen(attr)) \
 		return 2; \
+	strchop(buff); \
 	return 0; \
 }
 
@@ -144,6 +144,7 @@ declare_sysfs_get_str(cutype);
 declare_sysfs_get_str(vendor);
 declare_sysfs_get_str(model);
 declare_sysfs_get_str(rev);
+declare_sysfs_get_str(state);
 
 int
 sysfs_get_dev (struct sysfs_device * dev, char * buff, size_t len)
@@ -179,7 +180,7 @@ sysfs_get_size (struct sysfs_device * dev, unsigned long long * size)
 
 	return 0;
 }
-	
+
 int
 sysfs_get_fc_nodename (struct sysfs_device * dev, char * node,
 		       unsigned int host, unsigned int channel,
@@ -187,7 +188,7 @@ sysfs_get_fc_nodename (struct sysfs_device * dev, char * node,
 {
 	char attr_path[SYSFS_PATH_SIZE], *attr;
 
-	if (safe_sprintf(attr_path, 
+	if (safe_sprintf(attr_path,
 			 "/class/fc_transport/target%i:%i:%i",
 			 host, channel, target)) {
 		condlog(0, "attr_path too small");
@@ -202,10 +203,44 @@ sysfs_get_fc_nodename (struct sysfs_device * dev, char * node,
 
 	return 1;
 }
-	
-/*
- * udev might be slow creating node files : wait
- */
+
+int
+sysfs_set_scsi_tmo (struct multipath *mpp)
+{
+	char attr_path[SYSFS_PATH_SIZE];
+	struct path *pp;
+	int i;
+	char value[11];
+
+	if (!mpp->dev_loss && !mpp->fast_io_fail)
+		return 0;
+	vector_foreach_slot(mpp->paths, pp, i) {
+		if (safe_snprintf(attr_path, SYSFS_PATH_SIZE,
+				  "/class/fc_remote_ports/rport-%d:%d-%d",
+				  pp->sg_id.host_no, pp->sg_id.channel,
+				  pp->sg_id.scsi_id)) {
+			condlog(0, "attr_path '/class/fc_remote_ports/rport-%d:%d-%d' too large", pp->sg_id.host_no, pp->sg_id.channel, pp->sg_id.scsi_id);
+			return 1;
+		}
+		if (mpp->dev_loss){
+			snprintf(value, 11, "%u", mpp->dev_loss);
+			if (sysfs_attr_set_value(attr_path, "dev_loss_tmo",
+						 value))
+				return 1;
+		}
+		if (mpp->fast_io_fail){
+			if (mpp->fast_io_fail == -1)
+				sprintf(value, "off");
+			else
+				snprintf(value, 11, "%u", mpp->fast_io_fail);
+			if (sysfs_attr_set_value(attr_path, "fast_io_fail_tmo",
+						 value))
+				return 1;
+		}
+	}
+	return 0;
+}
+
 static int
 opennode (char * dev, int mode)
 {
@@ -235,27 +270,32 @@ devt2devname (char *devname, char *devt)
 	char block_path[FILE_NAME_SIZE];
 	struct stat statbuf;
 
+	memset(block_path, 0, FILE_NAME_SIZE);
 	if (sscanf(devt, "%u:%u", &major, &minor) != 2) {
 		condlog(0, "Invalid device number %s", devt);
 		return 1;
 	}
 
-	if ((fd = fopen("/proc/partitions", "r")) < 0) {
+	if (!(fd = fopen("/proc/partitions", "r"))) {
 		condlog(0, "Cannot open /proc/partitions");
 		return 1;
 	}
-	
+
 	while (!feof(fd)) {
 		int r = fscanf(fd,"%u %u %*d %s",&tmpmaj, &tmpmin, dev);
 		if (!r) {
-			fscanf(fd,"%*s\n");
+			r = fscanf(fd,"%*s\n");
 			continue;
 		}
 		if (r != 3)
 			continue;
 
 		if ((major == tmpmaj) && (minor == tmpmin)) {
-			sprintf(block_path, "/sys/block/%s", dev);
+			if (snprintf(block_path, FILE_NAME_SIZE, "/sys/block/%s", dev) >= FILE_NAME_SIZE) {
+				condlog(0, "device name %s is too long\n", dev);
+				fclose(fd);
+				return 1;
+			}
 			break;
 		}
 	}
@@ -273,71 +313,72 @@ devt2devname (char *devname, char *devt)
 		condlog(0, "sysfs entry %s is not a directory\n", block_path);
 		return 1;
 	}
+	basenamecpy(block_path, devname);
 	return 0;
 }
 
 int
 do_inq(int sg_fd, int cmddt, int evpd, unsigned int pg_op,
-       void *resp, int mx_resp_len, int noisy)
+       void *resp, int mx_resp_len)
 {
-        unsigned char inqCmdBlk[INQUIRY_CMDLEN] =
-            { INQUIRY_CMD, 0, 0, 0, 0, 0 };
-        unsigned char sense_b[SENSE_BUFF_LEN];
-        struct sg_io_hdr io_hdr;
+	unsigned char inqCmdBlk[INQUIRY_CMDLEN] =
+		{ INQUIRY_CMD, 0, 0, 0, 0, 0 };
+	unsigned char sense_b[SENSE_BUFF_LEN];
+	struct sg_io_hdr io_hdr;
 
-        if (cmddt)
-                inqCmdBlk[1] |= 2;
-        if (evpd)
-                inqCmdBlk[1] |= 1;
-        inqCmdBlk[2] = (unsigned char) pg_op;
+	if (cmddt)
+		inqCmdBlk[1] |= 2;
+	if (evpd)
+		inqCmdBlk[1] |= 1;
+	inqCmdBlk[2] = (unsigned char) pg_op;
 	inqCmdBlk[3] = (unsigned char)((mx_resp_len >> 8) & 0xff);
 	inqCmdBlk[4] = (unsigned char) (mx_resp_len & 0xff);
-        memset(&io_hdr, 0, sizeof (struct sg_io_hdr));
-        io_hdr.interface_id = 'S';
-        io_hdr.cmd_len = sizeof (inqCmdBlk);
-        io_hdr.mx_sb_len = sizeof (sense_b);
-        io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
-        io_hdr.dxfer_len = mx_resp_len;
-        io_hdr.dxferp = resp;
-        io_hdr.cmdp = inqCmdBlk;
-        io_hdr.sbp = sense_b;
-        io_hdr.timeout = DEF_TIMEOUT;
+	memset(&io_hdr, 0, sizeof (struct sg_io_hdr));
+	io_hdr.interface_id = 'S';
+	io_hdr.cmd_len = sizeof (inqCmdBlk);
+	io_hdr.mx_sb_len = sizeof (sense_b);
+	io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
+	io_hdr.dxfer_len = mx_resp_len;
+	io_hdr.dxferp = resp;
+	io_hdr.cmdp = inqCmdBlk;
+	io_hdr.sbp = sense_b;
+	io_hdr.timeout = DEF_TIMEOUT;
 
-        if (ioctl(sg_fd, SG_IO, &io_hdr) < 0)
-                return -1;
+	if (ioctl(sg_fd, SG_IO, &io_hdr) < 0)
+		return -1;
 
-        /* treat SG_ERR here to get rid of sg_err.[ch] */
-        io_hdr.status &= 0x7e;
-        if ((0 == io_hdr.status) && (0 == io_hdr.host_status) &&
-            (0 == io_hdr.driver_status))
-                return 0;
-        if ((SCSI_CHECK_CONDITION == io_hdr.status) ||
-            (SCSI_COMMAND_TERMINATED == io_hdr.status) ||
-            (SG_ERR_DRIVER_SENSE == (0xf & io_hdr.driver_status))) {
-                if (io_hdr.sbp && (io_hdr.sb_len_wr > 2)) {
-                        int sense_key;
-                        unsigned char * sense_buffer = io_hdr.sbp;
-                        if (sense_buffer[0] & 0x2)
-                                sense_key = sense_buffer[1] & 0xf;
-                        else
-                                sense_key = sense_buffer[2] & 0xf;
-                        if(RECOVERED_ERROR == sense_key)
-                                return 0;
-                }
-        }
-        return -1;
+	/* treat SG_ERR here to get rid of sg_err.[ch] */
+	io_hdr.status &= 0x7e;
+	if ((0 == io_hdr.status) && (0 == io_hdr.host_status) &&
+	    (0 == io_hdr.driver_status))
+		return 0;
+	if ((SCSI_CHECK_CONDITION == io_hdr.status) ||
+	    (SCSI_COMMAND_TERMINATED == io_hdr.status) ||
+	    (SG_ERR_DRIVER_SENSE == (0xf & io_hdr.driver_status))) {
+		if (io_hdr.sbp && (io_hdr.sb_len_wr > 2)) {
+			int sense_key;
+			unsigned char * sense_buffer = io_hdr.sbp;
+			if (sense_buffer[0] & 0x2)
+				sense_key = sense_buffer[1] & 0xf;
+			else
+				sense_key = sense_buffer[2] & 0xf;
+			if(RECOVERED_ERROR == sense_key)
+				return 0;
+		}
+	}
+	return -1;
 }
 
 static int
 get_serial (char * str, int maxlen, int fd)
 {
-        int len = 0;
-        char buff[MX_ALLOC_LEN + 1] = {0};
+	int len = 0;
+	char buff[MX_ALLOC_LEN + 1] = {0};
 
 	if (fd < 0)
-                return 1;
+		return 1;
 
-	if (0 == do_inq(fd, 0, 1, 0x80, buff, MX_ALLOC_LEN, 0)) {
+	if (0 == do_inq(fd, 0, 1, 0x80, buff, MX_ALLOC_LEN)) {
 		len = buff[3];
 		if (len >= maxlen)
 			return 1;
@@ -347,27 +388,73 @@ get_serial (char * str, int maxlen, int fd)
 		}
 		return 0;
 	}
-        return 1;
+	return 1;
 }
 
 static int
-get_inq (char * vendor, char * product, char * rev, int fd)
+get_inq (char * dev, char * vendor, char * product, char * rev, int fd)
 {
-	char buff[MX_ALLOC_LEN + 1] = {0};
+	unsigned char buff[MX_ALLOC_LEN + 1] = {0};
+	int len;
 
 	if (fd < 0)
 		return 1;
 
-	if (0 == do_inq(fd, 0, 0, 0, buff, MX_ALLOC_LEN, 0)) {
-		memcpy(vendor, buff + 8, 8);
-		vendor[8] = '\0';
-		memcpy(product, buff + 16, 16);
-		product[16] = '\0';
-		memcpy(rev, buff + 32, 4);
-		rev[4] = '\0';
-		return 0;
+	if (0 != do_inq(fd, 0, 0, 0, buff, MX_ALLOC_LEN))
+		return 1;
+
+	/* Check peripheral qualifier */
+	if ((buff[0] >> 5) != 0) {
+		int pqual = (buff[0] >> 5);
+		switch (pqual) {
+		case 1:
+			condlog(3, "%s: INQUIRY failed, LU not connected", dev);
+			break;
+		case 3:
+			condlog(3, "%s: INQUIRY failed, LU not supported", dev);
+			break;
+		default:
+			condlog(3, "%s: INQUIRY failed, Invalid PQ %x",
+				dev, pqual);
+			break;
+		}
+
+		return 1;
 	}
-	return 1;
+
+	len = buff[4] + 4;
+
+	if (len < 8) {
+		condlog(3, "%s: INQUIRY response too short (len %d)",
+			dev, len);
+		return 1;
+	}
+
+	len -= 8;
+	memset(vendor, 0x0, 8);
+	memcpy(vendor, buff + 8, len > 8 ? 8 : len);
+	vendor[8] = '\0';
+	strchop(vendor);
+	if (len <= 8)
+		return 0;
+
+	len -= 8;
+
+	memset(product, 0x0, 16);
+	memcpy(product, buff + 16, len > 16 ? 16 : len);
+	product[16] = '\0';
+	strchop(product);
+	if (len <= 16)
+		return 0;
+
+	len -= 16;
+
+	memset(rev, 0x0, 4);
+	memcpy(rev, buff + 32, 4);
+	rev[4] = '\0';
+	strchop(rev);
+
+	return 0;
 }
 
 static int
@@ -398,7 +485,7 @@ scsi_sysfs_pathinfo (struct path * pp, struct sysfs_device * parent)
 	/*
 	 * host / bus / target / lun
 	 */
-	basename(parent->devpath, attr_path);
+	basenamecpy(parent->devpath, attr_path);
 
 	sscanf(attr_path, "%i:%i:%i:%i",
 			&pp->sg_id.host_no,
@@ -456,8 +543,8 @@ ccw_sysfs_pathinfo (struct path * pp, struct sysfs_device * parent)
 
 	/*
 	 * host / bus / target / lun
-	 */	
-	basename(parent->devpath, attr_path);
+	 */
+	basenamecpy(parent->devpath, attr_path);
 	pp->sg_id.lun = 0;
 	sscanf(attr_path, "%i.%i.%x",
 			&pp->sg_id.host_no,
@@ -481,7 +568,7 @@ cciss_sysfs_pathinfo (struct path * pp, struct sysfs_device * dev)
 	/*
 	 * host / bus / target / lun
 	 */
-	basename(dev->devpath, attr_path);
+	basenamecpy(dev->devpath, attr_path);
 	pp->sg_id.lun = 0;
 	pp->sg_id.channel = 0;
 	sscanf(attr_path, "cciss!c%id%i",
@@ -528,6 +615,46 @@ struct sysfs_device *sysfs_device_from_path(struct path *pp)
 	return sysfs_device_get(sysdev);
 }
 
+int
+path_offline (struct path * pp)
+{
+	struct sysfs_device * parent;
+	char buff[SCSI_STATE_SIZE];
+
+	pp->sysdev = sysfs_device_from_path(pp);
+	if (!pp->sysdev) {
+		condlog(1, "%s: failed to get sysfs information", pp->dev);
+		return PATH_WILD;
+	}
+
+	parent = sysfs_device_get_parent(pp->sysdev);
+	if (!parent)
+		parent = pp->sysdev;
+	if (parent && !strncmp(parent->kernel, "block",5))
+		parent = sysfs_device_get_parent(parent);
+	if (!parent) {
+		condlog(1, "%s: failed to get parent", pp->dev);
+		return PATH_WILD;
+	}
+
+	if (sysfs_get_state(parent, buff, SCSI_STATE_SIZE))
+		return PATH_WILD;
+
+	condlog(3, "%s: state = %s", pp->dev, buff);
+
+	if (!strncmp(buff, "offline", 7)) {
+		pp->offline = 1;
+		return PATH_DOWN;
+	}
+	pp->offline = 0;
+	if (!strncmp(buff, "blocked", 7))
+		return PATH_PENDING;
+	else if (!strncmp(buff, "running", 7))
+		return PATH_UP;
+
+	return PATH_DOWN;
+}
+
 extern int
 sysfs_pathinfo(struct path * pp)
 {
@@ -545,6 +672,15 @@ sysfs_pathinfo(struct path * pp)
 	parent = sysfs_device_get_parent(pp->sysdev);
 	if (!parent)
 		parent = pp->sysdev;
+	if (parent && !strncmp(parent->kernel, "block",5))
+		parent = sysfs_device_get_parent(parent);
+	if (!parent) {
+		condlog(1, "%s: failed to get parent", pp->dev);
+		return 1;
+	}
+
+	if (!strncmp(pp->dev,"cciss",5))
+		strcpy(parent->subsystem,"cciss");
 
 	condlog(3, "%s: subsystem = %s", pp->dev, parent->subsystem);
 
@@ -552,7 +688,7 @@ sysfs_pathinfo(struct path * pp)
 		pp->bus = SYSFS_BUS_SCSI;
 	if (!strncmp(parent->subsystem, "ccw",3))
 		pp->bus = SYSFS_BUS_CCW;
-	if (!strncmp(pp->dev,"cciss",5))
+	if (!strncmp(parent->subsystem,"cciss",5))
 		pp->bus = SYSFS_BUS_CCISS;
 
 	if (pp->bus == SYSFS_BUS_UNDEF)
@@ -584,8 +720,14 @@ scsi_ioctl_pathinfo (struct path * pp, int mask)
 static int
 cciss_ioctl_pathinfo (struct path * pp, int mask)
 {
+	int ret;
+
 	if (mask & DI_SYSFS) {
-		get_inq(pp->vendor_id, pp->product_id, pp->rev, pp->fd);
+		ret = get_inq(pp->dev, pp->vendor_id, pp->product_id,
+			      pp->rev, pp->fd);
+		if (ret)
+			return ret;
+
 		condlog(3, "%s: vendor = %s", pp->dev, pp->vendor_id);
 		condlog(3, "%s: product = %s", pp->dev, pp->product_id);
 		condlog(3, "%s: revision = %s", pp->dev, pp->rev);
@@ -599,54 +741,64 @@ cciss_ioctl_pathinfo (struct path * pp, int mask)
 	return 0;
 }
 
-static int
-get_state (struct path * pp)
+int
+get_state (struct path * pp, int daemon)
 {
 	struct checker * c = &pp->checker;
+	int state;
 
-	if (!pp->mpp)
-		return 0;
+	condlog(3, "%s: get_state", pp->dev);
 
 	if (!checker_selected(c)) {
+		if (daemon)
+			pathinfo(pp, conf->hwtable, DI_SYSFS);
 		select_checker(pp);
-		if (!checker_selected(c))
-			return 1;
+		if (!checker_selected(c)) {
+			condlog(3, "%s: No checker selected", pp->dev);
+			return PATH_UNCHECKED;
+		}
 		checker_set_fd(c, pp->fd);
-		if (checker_init(c, &pp->mpp->mpcontext))
-			return 1;
+		if (checker_init(c, pp->mpp?&pp->mpp->mpcontext:NULL)) {
+			condlog(3, "%s: checker init failed", pp->dev);
+			return PATH_UNCHECKED;
+		}
 	}
-	pp->state = checker_check(c);
-	condlog(3, "%s: state = %i", pp->dev, pp->state);
-	if (pp->state == PATH_DOWN)
-		condlog(2, "%s: checker msg is \"%s\"",
+	state = path_offline(pp);
+	if (state != PATH_UP) {
+		condlog(3, "%s: path inaccessible", pp->dev);
+		return state;
+	}
+	if (daemon)
+		checker_set_async(c);
+	state = checker_check(c);
+	condlog(3, "%s: state = %i", pp->dev, state);
+	if (state == PATH_DOWN && strlen(checker_message(c)))
+		condlog(3, "%s: checker msg is \"%s\"",
 			pp->dev, checker_message(c));
-	return 0;
+	return state;
 }
 
 static int
 get_prio (struct path * pp)
 {
-	char buff[CALLOUT_MAX_SIZE];
-	char prio[16];
+	if (!pp)
+		return 0;
 
-	if (!pp->getprio_selected) {
-		select_getprio(pp);
-		pp->getprio_selected = 1;
+	if (!pp->prio) {
+		select_prio(pp);
+		if (!pp->prio) {
+			condlog(3, "%s: no prio selected", pp->dev);
+			return 1;
+		}
 	}
-	if (!pp->getprio) {
-		pp->priority = PRIO_DEFAULT;
-	} else if (apply_format(pp->getprio, &buff[0], pp)) {
-		condlog(0, "error formatting prio callout command");
+	pp->priority = prio_getprio(pp->prio, pp);
+	if (pp->priority < 0) {
+		condlog(3, "%s: %s prio error", pp->dev, prio_name(pp->prio));
 		pp->priority = PRIO_UNDEF;
 		return 1;
-	} else if (execute_program(buff, prio, 16)) {
-		condlog(0, "error calling out %s", buff);
-		pp->priority = PRIO_UNDEF;
-		return 1;
-	} else
-		pp->priority = atoi(prio);
-
-	condlog(3, "%s: prio = %u", pp->dev, pp->priority);
+	}
+	condlog(3, "%s: %s prio = %u",
+		pp->dev, prio_name(pp->prio), pp->priority);
 	return 0;
 }
 
@@ -654,6 +806,7 @@ static int
 get_uid (struct path * pp)
 {
 	char buff[CALLOUT_MAX_SIZE];
+	int i;
 
 	if (!pp->getuid)
 		select_getuid(pp);
@@ -662,10 +815,16 @@ get_uid (struct path * pp)
 		condlog(0, "error formatting uid callout command");
 		memset(pp->wwid, 0, WWID_SIZE);
 	} else if (execute_program(buff, pp->wwid, WWID_SIZE)) {
-		condlog(0, "error calling out %s", buff);
+		condlog(3, "error calling out %s", buff);
 		memset(pp->wwid, 0, WWID_SIZE);
 		return 1;
 	}
+	/* Strip any trailing blanks */
+	i = WWID_SIZE - 1;
+	while (i > 0 && pp->wwid[i] == ' ') {
+		pp->wwid[i] = '\0';
+		i--;
+	}
 	condlog(3, "%s: uid = %s (callout)", pp->dev ,pp->wwid);
 	return 0;
 }
@@ -701,8 +860,11 @@ pathinfo (struct path *pp, vector hwtable, int mask)
 	    cciss_ioctl_pathinfo(pp, mask))
 		goto blank;
 
-	if (mask & DI_CHECKER && get_state(pp))
-		goto blank;
+	if (mask & DI_CHECKER) {
+		pp->state = get_state(pp, 0);
+		if (pp->state == PATH_UNCHECKED || pp->state == PATH_WILD)
+			goto blank;
+	}
 
 	 /*
 	  * Retrieve path priority, even for PATH_DOWN paths if it has never
@@ -723,5 +885,6 @@ blank:
 	 */
 	memset(pp->wwid, 0, WWID_SIZE);
 	pp->state = PATH_DOWN;
+
 	return 0;
 }
