@@ -7,8 +7,7 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include <checkers.h>
-
+#include "checkers.h"
 #include "vector.h"
 #include "memory.h"
 #include "structs.h"
@@ -60,7 +59,7 @@ assemble_map (struct multipath * mp)
 	minio = mp->minio;
 	p = mp->params;
 	freechar = sizeof(mp->params);
-	
+
 	shift = snprintf(p, freechar, "%s %s %i %i",
 			 mp->features, mp->hwhandler,
 			 VECTOR_SIZE(mp->pg), mp->bestpg);
@@ -71,7 +70,7 @@ assemble_map (struct multipath * mp)
 	}
 	p += shift;
 	freechar -= shift;
-	
+
 	vector_foreach_slot (mp->pg, pgp, i) {
 		pgp = VECTOR_SLOT(mp->pg, i);
 		shift = snprintf(p, freechar, " %s %i 1", mp->selector,
@@ -146,6 +145,11 @@ disassemble_map (vector pathvec, char * params, struct multipath * mpp)
 			FREE(word);
 			return 1;
 		}
+		if ((mpp->no_path_retry == NO_PATH_RETRY_UNDEF) ||
+			(mpp->no_path_retry == NO_PATH_RETRY_FAIL) ||
+			(mpp->no_path_retry == NO_PATH_RETRY_QUEUE))
+			setup_feature(mpp, word);
+
 		FREE(word);
 	}
 
@@ -183,11 +187,14 @@ disassemble_map (vector pathvec, char * params, struct multipath * mpp)
 	num_pg = atoi(word);
 	FREE(word);
 
-	if (num_pg > 0 && !mpp->pg)
+	if (num_pg > 0 && !mpp->pg) {
 		mpp->pg = vector_alloc();
-	
-	if (!mpp->pg)
-		return 1;
+
+		if (!mpp->pg)
+			return 1;
+	} else
+		mpp->pg = NULL;
+
 	/*
 	 * first pg to try
 	 */
@@ -219,7 +226,7 @@ disassemble_map (vector pathvec, char * params, struct multipath * mpp)
 				goto out;
 
 			num_pg_args = atoi(word);
-			
+
 			if (merge_words(&mpp->selector, word, 1)) {
 				FREE(word);
 				goto out1;
@@ -237,7 +244,7 @@ disassemble_map (vector pathvec, char * params, struct multipath * mpp)
 		 * paths
 		 */
 		pgp = alloc_pathgroup();
-		
+
 		if (!pgp)
 			goto out;
 
@@ -278,10 +285,9 @@ disassemble_map (vector pathvec, char * params, struct multipath * mpp)
 
 				strncpy(pp->dev_t, word, BLK_DEV_SIZE);
 
-#ifndef DAEMON
-				if (store_path(pathvec, pp))
-					goto out;
-#endif
+				/* Only call this in multipath client mode */
+				if (!mpp->waiter && store_path(pathvec, pp))
+					goto out1;
 			}
 			FREE(word);
 
@@ -306,16 +312,22 @@ disassemble_map (vector pathvec, char * params, struct multipath * mpp)
 			pp->pgindex = i + 1;
 
 			for (k = 0; k < num_paths_args; k++)
-				if (k == 0 && !strncmp(mpp->selector,
-						       "round-robin", 11)) {
-					p += get_word(p, &word);
-					def_minio = atoi(word);
+				if (k == 0) {
+					if (!strncmp(mpp->selector,
+						     "round-robin", 11)) {
+						p += get_word(p, &word);
+						def_minio = atoi(word);
 
-					if (mpp->rr_weight == RR_WEIGHT_PRIO
-					    && pp->priority > 0)
-						def_minio /= pp->priority;
+						if (mpp->rr_weight == RR_WEIGHT_PRIO
+						    && pp->priority > 0)
+							def_minio /= pp->priority;
+
+						FREE(word);
+					} else {
+						p += get_word(p, NULL);
+						def_minio = 0;
+					}
 
-					FREE(word);
 					if (def_minio != mpp->minio)
 						mpp->minio = def_minio;
 				}
@@ -329,6 +341,7 @@ out1:
 	FREE(word);
 out:
 	free_pgvec(mpp->pg, KEEP_PATHS);
+	mpp->pg = NULL;
 	return 1;
 }
 
@@ -343,6 +356,7 @@ disassemble_status (char * params, struct multipath * mpp)
 	int num_pg;
 	int num_pg_args;
 	int num_paths;
+	int def_minio = 0;
 	struct path * pp;
 	struct pathgroup * pgp;
 
@@ -398,6 +412,9 @@ disassemble_status (char * params, struct multipath * mpp)
 	num_pg = atoi(word);
 	FREE(word);
 
+	if (num_pg == 0)
+		return 0;
+
 	/*
 	 * next pg to try
 	 */
@@ -433,7 +450,7 @@ disassemble_status (char * params, struct multipath * mpp)
 		FREE(word);
 
 		/*
-		 * undef ?
+		 * PG Status (discarded, would be '0' anyway)
 		 */
 		p += get_word(p, NULL);
 
@@ -496,8 +513,17 @@ disassemble_status (char * params, struct multipath * mpp)
 			/*
 			 * selector args
 			 */
-			for (k = 0; k < num_pg_args; k++)
-				p += get_word(p, NULL);
+			for (k = 0; k < num_pg_args; k++) {
+				if (!strncmp(mpp->selector,
+					     "least-pending", 13)) {
+					p += get_word(p, &word);
+					if (sscanf(word,"%d:*d",
+						   &def_minio) == 1 &&
+					    def_minio != mpp->minio)
+							mpp->minio = def_minio;
+				} else
+					p += get_word(p, NULL);
+			}
 		}
 	}
 	return 0;
