@@ -1,11 +1,11 @@
-/* 
+/*
  * Part:        Configuration file parser/reader. Place into the dynamic
  *              data structure representation the conf file
- *  
+ *
  * Version:     $Id: parser.c,v 1.0.3 2003/05/11 02:28:03 acassen Exp $
- * 
+ *
  * Author:      Alexandre Cassen, <acassen@linux-vs.org>
- *              
+ *
  *              This program is distributed in the hope that it will be useful,
  *              but WITHOUT ANY WARRANTY; without even the implied warranty of
  *              MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
@@ -43,7 +43,7 @@ keyword_alloc(vector keywords, char *string, int (*handler) (vector),
 
 	if (!keyword)
 		return 1;
-	
+
 	if (!vector_alloc_slot(keywords)) {
 		FREE(keyword);
 		return 1;
@@ -172,7 +172,7 @@ snprint_keyword(char *buff, int len, char *fmt, struct keyword *kw, void *data)
 		f++;
 		switch(*f) {
 		case 'k':
-			fwd += snprintf(buff + fwd, len - fwd, kw->string);
+			fwd += snprintf(buff + fwd, len - fwd, "%s", kw->string);
 			break;
 		case 'v':
 			r = kw->print(buff + fwd, len - fwd, data);
@@ -203,7 +203,7 @@ alloc_strvec(char *string)
 	cp = string;
 
 	/* Skip white spaces */
-	while (isspace((int) *cp) && *cp != '\0')
+	while ((isspace((int) *cp) || !isascii((int) *cp)) && *cp != '\0')
 		cp++;
 
 	/* Return if there is only white spaces */
@@ -239,10 +239,20 @@ alloc_strvec(char *string)
 				in_string = 0;
 			else
 				in_string = 1;
+		} else if (!in_string && (*cp == '{' || *cp == '}')) {
+			token = MALLOC(2);
 
+			if (!token)
+				goto out;
+
+			*(token) = *cp;
+			*(token + 1) = '\0';
+			cp++;
 		} else {
-			while ((in_string || !isspace((int) *cp)) && *cp
-				!= '\0' && *cp != '"')
+			while ((in_string ||
+				(!isspace((int) *cp) && isascii((int) *cp) &&
+				 *cp != '!' && *cp != '#' && *cp != '{' &&
+				 *cp != '}')) && *cp != '\0' && *cp != '"')
 				cp++;
 			strlen = cp - start;
 			token = MALLOC(strlen + 1);
@@ -255,7 +265,8 @@ alloc_strvec(char *string)
 		}
 		vector_set_slot(strvec, token);
 
-		while (isspace((int) *cp) && *cp != '\0')
+		while ((isspace((int) *cp) || !isascii((int) *cp))
+		       && *cp != '\0')
 			cp++;
 		if (*cp == '\0' || *cp == '!' || *cp == '#')
 			return strvec;
@@ -292,14 +303,14 @@ read_value_block(void)
 	vector vec = NULL;
 	vector elements = vector_alloc();
 
+	if (!elements)
+		return NULL;
+
 	buf = (char *) MALLOC(MAXBUF);
 
 	if (!buf)
 		return NULL;
 
-	if (!elements)
-		goto out;
-
 	while (read_line(buf, MAXBUF)) {
 		vec = alloc_strvec(buf);
 		if (vec) {
@@ -313,10 +324,14 @@ read_value_block(void)
 				for (i = 0; i < VECTOR_SIZE(vec); i++) {
 					str = VECTOR_SLOT(vec, i);
 					dup = (char *) MALLOC(strlen(str) + 1);
+					if (!dup)
+						goto out;
 					memcpy(dup, str, strlen(str));
 
-					if (!vector_alloc_slot(elements))
+					if (!vector_alloc_slot(elements)) {
+						free_strvec(vec);
 						goto out1;
+					}
 
 					vector_set_slot(elements, dup);
 				}
@@ -387,16 +402,17 @@ set_value(vector strvec)
 				alloc =
 				    REALLOC(alloc, sizeof (char *) * (len + 1));
 				tmp = VECTOR_SLOT(strvec, i-1);
-				if (*str != '"' && *tmp != '"')
+				if (alloc && *str != '"' && *tmp != '"')
 					strncat(alloc, " ", 1);
 			}
 
-			if (i != VECTOR_SIZE(strvec)-1)
+			if (alloc && i != VECTOR_SIZE(strvec)-1)
 				strncat(alloc, str, strlen(str));
 		}
 	} else {
 		alloc = MALLOC(sizeof (char *) * (size + 1));
-		memcpy(alloc, str, size);
+		if (alloc)
+			memcpy(alloc, str, size);
 	}
 	return alloc;
 }
@@ -447,7 +463,7 @@ process_stream(vector keywords)
 				break;
 			}
 		}
-		
+
 		free_strvec(strvec);
 	}
 
