@@ -5,8 +5,7 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include <checkers.h>
-
+#include "checkers.h"
 #include "util.h"
 #include "memory.h"
 #include "vector.h"
@@ -68,7 +67,7 @@ group_by_node_name (struct multipath * mp) {
 	struct path * pp;
 	struct pathgroup * pgp;
 	struct path * pp2;
-	
+
 	if (!mp->pg)
 		mp->pg = vector_alloc();
 
@@ -104,12 +103,12 @@ group_by_node_name (struct multipath * mp) {
 		bitmap[i] = 1;
 
 		for (j = i + 1; j < VECTOR_SIZE(mp->paths); j++) {
-			
+
 			if (bitmap[j])
 				continue;
 
 			pp2 = VECTOR_SLOT(mp->paths, j);
-			
+
 			if (!strncmp(pp->tgt_node_name, pp2->tgt_node_name,
 					NODE_NAME_SIZE)) {
 				if (store_path(pgp->paths, pp2))
@@ -127,6 +126,7 @@ out1:
 	FREE(bitmap);
 out:
 	free_pgvec(mp->pg, KEEP_PATHS);
+	mp->pg = NULL;
 	return 1;
 }
 
@@ -140,7 +140,7 @@ group_by_serial (struct multipath * mp) {
 	struct path * pp;
 	struct pathgroup * pgp;
 	struct path * pp2;
-	
+
 	if (!mp->pg)
 		mp->pg = vector_alloc();
 
@@ -172,16 +172,16 @@ group_by_serial (struct multipath * mp) {
 		/* feed the first path */
 		if (store_path(pgp->paths, pp))
 			goto out1;
-				
+
 		bitmap[i] = 1;
 
 		for (j = i + 1; j < VECTOR_SIZE(mp->paths); j++) {
-			
+
 			if (bitmap[j])
 				continue;
 
 			pp2 = VECTOR_SLOT(mp->paths, j);
-			
+
 			if (0 == strcmp(pp->serial, pp2->serial)) {
 				if (store_path(pgp->paths, pp2))
 					goto out1;
@@ -198,6 +198,7 @@ out1:
 	FREE(bitmap);
 out:
 	free_pgvec(mp->pg, KEEP_PATHS);
+	mp->pg = NULL;
 	return 1;
 }
 
@@ -210,7 +211,7 @@ one_path_per_group (struct multipath * mp)
 
 	if (!mp->pg)
 		mp->pg = vector_alloc();
-	
+
 	if (!mp->pg)
 		return 1;
 
@@ -223,7 +224,7 @@ one_path_per_group (struct multipath * mp)
 
 		if (store_pathgroup(mp->pg, pgp))
 			goto out;
-		
+
 		if (store_path(pgp->paths, pp))
 			goto out;
 	}
@@ -232,6 +233,7 @@ one_path_per_group (struct multipath * mp)
 	return 0;
 out:
 	free_pgvec(mp->pg, KEEP_PATHS);
+	mp->pg = NULL;
 	return 1;
 }
 
@@ -249,21 +251,24 @@ one_group (struct multipath * mp)	/* aka multibus */
 	if (!mp->pg)
 		return 1;
 
-	pgp = alloc_pathgroup();
-	
-	if (!pgp)
-		goto out;
-	
-	vector_free(pgp->paths);
-	pgp->paths = mp->paths;
-	mp->paths = NULL;
+	if (VECTOR_SIZE(mp->paths) > 0) {
+		pgp = alloc_pathgroup();
 
-	if (store_pathgroup(mp->pg, pgp))
-		goto out;
+		if (!pgp)
+			goto out;
+
+		vector_free(pgp->paths);
+		pgp->paths = mp->paths;
+		mp->paths = NULL;
+
+		if (store_pathgroup(mp->pg, pgp))
+			goto out;
+	}
 
 	return 0;
 out:
 	free_pgvec(mp->pg, KEEP_PATHS);
+	mp->pg = NULL;
 	return 1;
 }
 
@@ -303,7 +308,7 @@ group_by_prio (struct multipath * mp)
 
 		if (!pgp)
 			goto out;
-		
+
 		if (store_path(pgp->paths, VECTOR_SLOT(mp->paths, 0)))
 				goto out;
 
@@ -338,6 +343,7 @@ group_by_prio (struct multipath * mp)
 	return 0;
 out:
 	free_pgvec(mp->pg, KEEP_PATHS);
+	mp->pg = NULL;
 	return 1;
 
 }
