@@ -7,9 +7,9 @@
 #include <stdarg.h>
 #include <sys/stat.h>
 #include <dirent.h>
+#include <unistd.h>
 
-#include <checkers.h>
-
+#include "checkers.h"
 #include "vector.h"
 #include "structs.h"
 #include "structs_vec.h"
@@ -21,14 +21,20 @@
 #include "defaults.h"
 #include "parser.h"
 #include "blacklist.h"
+#include "switchgroup.h"
+#include "devmapper.h"
+#include "uevent.h"
 
 #define MAX(x,y) (x > y) ? x : y
 #define TAIL     (line + len - 1 - c)
 #define NOPAD    s = c
 #define PAD(x)   while ((int)(c - s) < (x) && (c < (line + len - 1))) \
 			*c++ = ' '; s = c
+#define ENDLINE \
+		if (c > line) \
+			line[c - line - 1] = '\n'
 #define PRINT(var, size, format, args...)      \
-	         fwd = snprintf(var, size, format, ##args); \
+		fwd = snprintf(var, size, format, ##args); \
 		 c += (fwd >= size) ? size : fwd;
 
 /*
@@ -59,7 +65,7 @@ snprint_size (char * buff, size_t len, unsigned long long size)
 	char fmt[6] = {};
 	char units[] = {'K','M','G','T','P'};
 	char *u = units;
-	
+
 	while (s >= 1024 && *u != 'P') {
 		s = s / 1024;
 		u++;
@@ -89,26 +95,40 @@ snprint_sysfs (char * buff, size_t len, struct multipath * mpp)
 {
 	if (mpp->dmi)
 		return snprintf(buff, len, "dm-%i", mpp->dmi->minor);
+	else
+		return snprintf(buff, len, "undef");
+}
 
-	return 0;
+static int
+snprint_ro (char * buff, size_t len, struct multipath * mpp)
+{
+	if (!mpp->dmi)
+		return snprintf(buff, len, "undef");
+	if (mpp->dmi->read_only)
+		return snprintf(buff, len, "ro");
+	else
+		return snprintf(buff, len, "rw");
 }
 
 static int
 snprint_progress (char * buff, size_t len, int cur, int total)
 {
-	int i = PROGRESS_LEN * cur / total;
-	int j = PROGRESS_LEN - i;
 	char * c = buff;
 	char * end = buff + len;
-	
-	while (i-- > 0) {
-		c += snprintf(c, len, "X");
-		if ((len = (end - c)) <= 1) goto out;
-	}
 
-	while (j-- > 0) {
-		c += snprintf(c, len,  ".");
-		if ((len = (end - c)) <= 1) goto out;
+	if (total > 0) {
+		int i = PROGRESS_LEN * cur / total;
+		int j = PROGRESS_LEN - i;
+
+		while (i-- > 0) {
+			c += snprintf(c, len, "X");
+			if ((len = (end - c)) <= 1) goto out;
+		}
+
+		while (j-- > 0) {
+			c += snprintf(c, len,  ".");
+			if ((len = (end - c)) <= 1) goto out;
+		}
 	}
 
 	c += snprintf(c, len, " %i/%i", cur, total);
@@ -117,7 +137,7 @@ out:
 	buff[c - buff + 1] = '\0';
 	return (c - buff + 1);
 }
-	
+
 static int
 snprint_failback (char * buff, size_t len, struct multipath * mpp)
 {
@@ -220,6 +240,16 @@ snprint_multipath_uuid (char * buff, size_t len, struct multipath * mpp)
 	return snprint_str(buff, len, mpp->wwid);
 }
 
+static int
+snprint_multipath_vpr (char * buff, size_t len, struct multipath * mpp)
+{
+	struct path * pp = first_path(mpp);
+	if (!pp)
+		return 0;
+	return snprintf(buff, len, "%s,%s",
+			pp->vendor_id, pp->product_id);
+}
+
 static int
 snprint_action (char * buff, size_t len, struct multipath * mpp)
 {
@@ -279,20 +309,29 @@ snprint_dev_t (char * buff, size_t len, struct path * pp)
 		return snprint_str(buff, len, pp->dev_t);
 }
 
+static int
+snprint_offline (char * buff, size_t len, struct path * pp)
+{
+	if (pp->offline)
+		return snprintf(buff, len, "offline");
+	else
+		return snprintf(buff, len, "running");
+}
+
 static int
 snprint_chk_state (char * buff, size_t len, struct path * pp)
 {
 	switch (pp->state) {
 	case PATH_UP:
-		return snprintf(buff, len, "[ready]");
+		return snprintf(buff, len, "ready");
 	case PATH_DOWN:
-		return snprintf(buff, len, "[faulty]");
+		return snprintf(buff, len, "faulty");
 	case PATH_SHAKY:
-		return snprintf(buff, len, "[shaky]");
+		return snprintf(buff, len, "shaky");
 	case PATH_GHOST:
-		return snprintf(buff, len, "[ghost]");
+		return snprintf(buff, len, "ghost");
 	default:
-		return snprintf(buff, len, "[undef]");
+		return snprintf(buff, len, "undef");
 	}
 }
 
@@ -301,11 +340,11 @@ snprint_dm_path_state (char * buff, size_t len, struct path * pp)
 {
 	switch (pp->dmstate) {
 	case PSTATE_ACTIVE:
-		return snprintf(buff, len, "[active]");
+		return snprintf(buff, len, "active");
 	case PSTATE_FAILED:
-		return snprintf(buff, len, "[failed]");
+		return snprintf(buff, len, "failed");
 	default:
-		return snprintf(buff, len, "[undef]");
+		return snprintf(buff, len, "undef");
 	}
 }
 
@@ -313,14 +352,14 @@ static int
 snprint_vpr (char * buff, size_t len, struct path * pp)
 {
 	return snprintf(buff, len, "%s,%s",
-		        pp->vendor_id, pp->product_id);
+			pp->vendor_id, pp->product_id);
 }
 
 static int
 snprint_next_check (char * buff, size_t len, struct path * pp)
 {
 	if (!pp->mpp)
-		return snprintf(buff, len, "[orphan]");
+		return snprintf(buff, len, "orphan");
 
 	return snprint_progress(buff, len, pp->tick, pp->checkint);
 }
@@ -340,7 +379,17 @@ snprint_pg_selector (char * buff, size_t len, struct pathgroup * pgp)
 static int
 snprint_pg_pri (char * buff, size_t len, struct pathgroup * pgp)
 {
-	return snprint_int(buff, len, pgp->priority);
+	int avg_priority = 0;
+	/*
+	 * path group priority is not updated for every path prio change,
+	 * but only on switch group code path.
+	 *
+	 * Printing is another reason to update.
+	 */
+	path_group_prio_update(pgp);
+	if (pgp->enabled_paths)
+		avg_priority = pgp->priority / pgp->enabled_paths;
+	return snprint_int(buff, len, avg_priority);
 }
 
 static int
@@ -348,13 +397,13 @@ snprint_pg_state (char * buff, size_t len, struct pathgroup * pgp)
 {
 	switch (pgp->status) {
 	case PGSTATE_ENABLED:
-		return snprintf(buff, len, "[enabled]");
+		return snprintf(buff, len, "enabled");
 	case PGSTATE_DISABLED:
-		return snprintf(buff, len, "[disabled]");
+		return snprintf(buff, len, "disabled");
 	case PGSTATE_ACTIVE:
-		return snprintf(buff, len, "[active]");
+		return snprintf(buff, len, "active");
 	default:
-		return snprintf(buff, len, "[undef]");
+		return snprintf(buff, len, "undef");
 	}
 }
 
@@ -364,6 +413,19 @@ snprint_path_size (char * buff, size_t len, struct path * pp)
 	return snprint_size(buff, len, pp->size);
 }
 
+static int
+snprint_path_serial (char * buff, size_t len, struct path * pp)
+{
+	return snprint_str(buff, len, pp->serial);
+}
+
+static int
+snprint_path_checker (char * buff, size_t len, struct path * pp)
+{
+	struct checker * c = &pp->checker;
+	return snprint_str(buff, len, c->name);
+}
+
 struct multipath_data mpd[] = {
 	{'n', "name",          0, snprint_name},
 	{'w', "uuid",          0, snprint_multipath_uuid},
@@ -371,6 +433,7 @@ struct multipath_data mpd[] = {
 	{'F', "failback",      0, snprint_failback},
 	{'Q', "queueing",      0, snprint_queueing},
 	{'N', "paths",         0, snprint_nb_paths},
+	{'r', "write_prot",    0, snprint_ro},
 	{'t', "dm-st",         0, snprint_dm_map_state},
 	{'S', "size",          0, snprint_multipath_size},
 	{'f', "features",      0, snprint_features},
@@ -381,6 +444,7 @@ struct multipath_data mpd[] = {
 	{'2', "map_loads",     0, snprint_map_loads},
 	{'3', "total_q_time",  0, snprint_total_q_time},
 	{'4', "q_timeouts",    0, snprint_q_timeouts},
+	{'s', "vend/prod/rev", 0, snprint_multipath_vpr},
 	{0, NULL, 0 , NULL}
 };
 
@@ -390,11 +454,14 @@ struct path_data pd[] = {
 	{'d', "dev",           0, snprint_dev},
 	{'D', "dev_t",         0, snprint_dev_t},
 	{'t', "dm_st",         0, snprint_dm_path_state},
+	{'o', "dev_st",        0, snprint_offline},
 	{'T', "chk_st",        0, snprint_chk_state},
 	{'s', "vend/prod/rev", 0, snprint_vpr},
+	{'c', "checker",       0, snprint_path_checker},
 	{'C', "next_check",    0, snprint_next_check},
 	{'p', "pri",           0, snprint_pri},
 	{'S', "size",          0, snprint_path_size},
+	{'z', "serial",        0, snprint_path_serial},
 	{0, NULL, 0 , NULL}
 };
 
@@ -402,17 +469,41 @@ struct pathgroup_data pgd[] = {
 	{'s', "selector",      0, snprint_pg_selector},
 	{'p', "pri",           0, snprint_pg_pri},
 	{'t', "dm_st",         0, snprint_pg_state},
+	{0, NULL, 0 , NULL}
 };
 
+int
+snprint_wildcards (char * buff, int len)
+{
+	int i, fwd = 0;
+
+	fwd += snprintf(buff + fwd, len - fwd, "multipath format wildcards:\n");
+	for (i = 0; mpd[i].header; i++)
+		fwd += snprintf(buff + fwd, len - fwd, "%%%c  %s\n",
+				mpd[i].wildcard, mpd[i].header);
+	fwd += snprintf(buff + fwd, len - fwd, "\npath format wildcards:\n");
+	for (i = 0; pd[i].header; i++)
+		fwd += snprintf(buff + fwd, len - fwd, "%%%c  %s\n",
+				pd[i].wildcard, pd[i].header);
+	fwd += snprintf(buff + fwd, len - fwd, "\npathgroup format wildcards:\n");
+	for (i = 0; pgd[i].header; i++)
+		fwd += snprintf(buff + fwd, len - fwd, "%%%c  %s\n",
+				pgd[i].wildcard, pgd[i].header);
+	return fwd;
+}
+
 void
-get_path_layout (vector pathvec)
+get_path_layout (vector pathvec, int header)
 {
 	int i, j;
 	char buff[MAX_FIELD_LEN];
 	struct path * pp;
 
 	for (j = 0; pd[j].header; j++) {
-		pd[j].width = strlen(pd[j].header);
+		if (header)
+			pd[j].width = strlen(pd[j].header);
+		else
+			pd[j].width = 0;
 
 		vector_foreach_slot (pathvec, pp, i) {
 			pd[j].snprint(buff, MAX_FIELD_LEN, pp);
@@ -421,15 +512,27 @@ get_path_layout (vector pathvec)
 	}
 }
 
+static void
+reset_multipath_layout (void)
+{
+	int i;
+
+	for (i = 0; mpd[i].header; i++)
+		mpd[i].width = 0;
+}
+
 void
-get_multipath_layout (vector mpvec)
+get_multipath_layout (vector mpvec, int header)
 {
 	int i, j;
 	char buff[MAX_FIELD_LEN];
 	struct multipath * mpp;
 
 	for (j = 0; mpd[j].header; j++) {
-		mpd[j].width = strlen(mpd[j].header);
+		if (header)
+			mpd[j].width = strlen(mpd[j].header);
+		else
+			mpd[j].width = 0;
 
 		vector_foreach_slot (mpvec, mpp, i) {
 			mpd[j].snprint(buff, MAX_FIELD_LEN, mpp);
@@ -483,6 +586,8 @@ snprint_multipath_header (char * line, int len, char * format)
 	int fwd;
 	struct multipath_data * data;
 
+	memset(line, 0, len);
+
 	do {
 		if (!TAIL)
 			break;
@@ -493,17 +598,15 @@ snprint_multipath_header (char * line, int len, char * format)
 			continue;
 		}
 		f++;
-		
+
 		if (!(data = mpd_lookup(*f)))
-			break; /* unknown wildcard */
-		
-		PRINT(c, TAIL, data->header);
+			continue; /* unknown wildcard */
+
+		PRINT(c, TAIL, "%s", data->header);
 		PAD(data->width);
 	} while (*f++);
 
-	line[c - line - 1] = '\n';
-	line[c - line] = '\0';
-
+	ENDLINE;
 	return (c - line);
 }
 
@@ -518,6 +621,8 @@ snprint_multipath (char * line, int len, char * format,
 	struct multipath_data * data;
 	char buff[MAX_FIELD_LEN] = {};
 
+	memset(line, 0, len);
+
 	do {
 		if (!TAIL)
 			break;
@@ -528,19 +633,17 @@ snprint_multipath (char * line, int len, char * format,
 			continue;
 		}
 		f++;
-		
+
 		if (!(data = mpd_lookup(*f)))
-			break;
-		
+			continue;
+
 		data->snprint(buff, MAX_FIELD_LEN, mpp);
-		PRINT(c, TAIL, buff);
+		PRINT(c, TAIL, "%s", buff);
 		PAD(data->width);
 		buff[0] = '\0';
 	} while (*f++);
 
-	line[c - line - 1] = '\n';
-	line[c - line] = '\0';
-
+	ENDLINE;
 	return (c - line);
 }
 
@@ -553,6 +656,8 @@ snprint_path_header (char * line, int len, char * format)
 	int fwd;
 	struct path_data * data;
 
+	memset(line, 0, len);
+
 	do {
 		if (!TAIL)
 			break;
@@ -563,17 +668,15 @@ snprint_path_header (char * line, int len, char * format)
 			continue;
 		}
 		f++;
-		
+
 		if (!(data = pd_lookup(*f)))
-			break; /* unknown wildcard */
-		
-		PRINT(c, TAIL, data->header);
+			continue; /* unknown wildcard */
+
+		PRINT(c, TAIL, "%s", data->header);
 		PAD(data->width);
 	} while (*f++);
 
-	line[c - line - 1] = '\n';
-	line[c - line] = '\0';
-
+	ENDLINE;
 	return (c - line);
 }
 
@@ -588,6 +691,8 @@ snprint_path (char * line, int len, char * format,
 	struct path_data * data;
 	char buff[MAX_FIELD_LEN];
 
+	memset(line, 0, len);
+
 	do {
 		if (!TAIL)
 			break;
@@ -598,18 +703,16 @@ snprint_path (char * line, int len, char * format,
 			continue;
 		}
 		f++;
-		
+
 		if (!(data = pd_lookup(*f)))
-			break;
-		
+			continue;
+
 		data->snprint(buff, MAX_FIELD_LEN, pp);
-		PRINT(c, TAIL, buff);
+		PRINT(c, TAIL, "%s", buff);
 		PAD(data->width);
 	} while (*f++);
 
-	line[c - line - 1] = '\n';
-	line[c - line] = '\0';
-
+	ENDLINE;
 	return (c - line);
 }
 
@@ -624,6 +727,8 @@ snprint_pathgroup (char * line, int len, char * format,
 	struct pathgroup_data * data;
 	char buff[MAX_FIELD_LEN];
 
+	memset(line, 0, len);
+
 	do {
 		if (!TAIL)
 			break;
@@ -634,18 +739,16 @@ snprint_pathgroup (char * line, int len, char * format,
 			continue;
 		}
 		f++;
-		
+
 		if (!(data = pgd_lookup(*f)))
-			break;
-		
+			continue;
+
 		data->snprint(buff, MAX_FIELD_LEN, pgp);
-		PRINT(c, TAIL, buff);
+		PRINT(c, TAIL, "%s", buff);
 		PAD(data->width);
 	} while (*f++);
 
-	line[c - line - 1] = '\n';
-	line[c - line] = '\0';
-
+	ENDLINE;
 	return (c - line);
 }
 
@@ -654,6 +757,7 @@ print_multipath_topology (struct multipath * mpp, int verbosity)
 {
 	char buff[MAX_LINE_LEN * MAX_LINES] = {};
 
+	memset(&buff[0], 0, MAX_LINE_LEN * MAX_LINES);
 	snprint_multipath_topology(&buff[0], MAX_LINE_LEN * MAX_LINES,
 				   mpp, verbosity);
 	printf("%s", buff);
@@ -668,31 +772,38 @@ snprint_multipath_topology (char * buff, int len, struct multipath * mpp,
 	struct pathgroup * pgp = NULL;
 	char style[64];
 	char * c = style;
+	char fmt[64];
+	char * f;
 
 	if (verbosity <= 0)
 		return fwd;
 
+	reset_multipath_layout();
+
 	if (verbosity == 1)
 		return snprint_multipath(buff, len, "%n", mpp);
 
+	if(isatty(1))
+		c += sprintf(c, "%c[%dm", 0x1B, 1); /* bold on */
+
 	if (verbosity > 1 &&
 	    mpp->action != ACT_NOTHING &&
 	    mpp->action != ACT_UNDEF)
 			c += sprintf(c, "%%A: ");
 
 	c += sprintf(c, "%%n");
-	
-	if (strncmp(mpp->alias, mpp->wwid, WWID_SIZE))
-		c += sprintf(c, " (%%w) ");
 
-	c += sprintf(c, "%%d ");
-	c += snprint_vpr(c, 24, first_path(mpp));
+	if (strncmp(mpp->alias, mpp->wwid, WWID_SIZE))
+		c += sprintf(c, " (%%w)");
+
+	c += sprintf(c, " %%d %%s");
+	if(isatty(1))
+		c += sprintf(c, "%c[%dm", 0x1B, 0); /* bold off */
 
 	fwd += snprint_multipath(buff + fwd, len - fwd, style, mpp);
 	if (fwd > len)
 		return len;
-	fwd += snprint_multipath(buff + fwd, len - fwd,
-				 "[size=%S][features=%f][hwhandler=%h]", mpp);
+	fwd += snprint_multipath(buff + fwd, len - fwd, PRINT_MAP_PROPS, mpp);
 	if (fwd > len)
 		return len;
 
@@ -700,15 +811,26 @@ snprint_multipath_topology (char * buff, int len, struct multipath * mpp,
 		return fwd;
 
 	vector_foreach_slot (mpp->pg, pgp, j) {
+		f=fmt;
 		pgp->selector = mpp->selector; /* hack */
-		fwd += snprint_pathgroup(buff + fwd, len - fwd,
-					 PRINT_PG_INDENT, pgp);
+		if (j + 1 < VECTOR_SIZE(mpp->pg)) {
+			strcpy(f, "|-+- " PRINT_PG_INDENT);
+		} else
+			strcpy(f, "`-+- " PRINT_PG_INDENT);
+		fwd += snprint_pathgroup(buff + fwd, len - fwd, fmt, pgp);
 		if (fwd > len)
 			return len;
 
 		vector_foreach_slot (pgp->paths, pp, i) {
-			fwd += snprint_path(buff + fwd, len - fwd,
-					    PRINT_PATH_INDENT, pp);
+			f=fmt;
+			if (*f != '|')
+				*f=' ';
+			f++;
+			if (i + 1 < VECTOR_SIZE(pgp->paths))
+				strcpy(f, " |- " PRINT_PATH_INDENT);
+			else
+				strcpy(f, " `- " PRINT_PATH_INDENT);
+			fwd += snprint_path(buff + fwd, len - fwd, fmt, pp);
 			if (fwd > len)
 				return len;
 		}
@@ -854,7 +976,6 @@ snprint_defaults (char * buff, int len)
 	if (fwd > len)
 		return len;
 	return fwd;
-	
 }
 
 static int
@@ -920,7 +1041,7 @@ snprint_blacklist_report (char * buff, int len)
 	if ((len - fwd - threshold) <= 0)
 		return len;
 	fwd += snprintf(buff + fwd, len - fwd, "device node rules:\n"
-			                       "- blacklist:\n");
+					       "- blacklist:\n");
 	if (!snprint_blacklist_group(buff, len, &fwd, &conf->blist_devnode))
 		return len;
 
@@ -933,7 +1054,7 @@ snprint_blacklist_report (char * buff, int len)
 	if ((len - fwd - threshold) <= 0)
 		return len;
 	fwd += snprintf(buff + fwd, len - fwd, "wwid rules:\n"
-			                       "- blacklist:\n");
+					       "- blacklist:\n");
 	if (snprint_blacklist_group(buff, len, &fwd, &conf->blist_wwid) == 0)
 		return len;
 
@@ -946,7 +1067,7 @@ snprint_blacklist_report (char * buff, int len)
 	if ((len - fwd - threshold) <= 0)
 		return len;
 	fwd += snprintf(buff + fwd, len - fwd, "device rules:\n"
-			                       "- blacklist:\n");
+					       "- blacklist:\n");
 	if (snprint_blacklist_devgroup(buff, len, &fwd, &conf->blist_device) == 0)
 		return len;
 
@@ -1097,6 +1218,38 @@ snprint_blacklist_except (char * buff, int len)
 	return fwd;
 }
 
+extern int
+snprint_status (char * buff, int len, struct vectors *vecs)
+{
+	int fwd = 0;
+	int i;
+	unsigned int count[PATH_MAX_STATE] = {0};
+	struct path * pp;
+
+	vector_foreach_slot (vecs->pathvec, pp, i) {
+		count[pp->state]++;
+	}
+	fwd += snprintf(buff + fwd, len - fwd, "path checker states:\n");
+	for (i=0; i<PATH_MAX_STATE; i++) {
+		if (!count[i])
+			continue;
+		fwd += snprintf(buff + fwd, len - fwd, "%-20s%u\n",
+				checker_state_name(i), count[i]);
+	}
+
+        int monitored_count = 0;
+
+        vector_foreach_slot(vecs->pathvec, pp, i)
+                if (pp->fd != -1)
+                        monitored_count++;
+        fwd += snprintf(buff + fwd, len - fwd, "\npaths: %d\nbusy: %s\n",
+			monitored_count, is_uevent_busy()? "True" : "False");
+
+	if (fwd > len)
+		return len;
+	return fwd;
+}
+
 extern int
 snprint_devices (char * buff, int len, struct vectors *vecs)
 {
@@ -1118,15 +1271,16 @@ snprint_devices (char * buff, int len, struct vectors *vecs)
 		return len;
 	fwd += snprintf(buff + fwd, len - fwd, "available block devices:\n");
 
-	strcpy(devpath,"/sys/block");
-	devptr = devpath + 10;
+	strcpy(devpath,"/sys/block/");
 	while ((blkdev = readdir(blkdir)) != NULL) {
 		if ((strcmp(blkdev->d_name,".") == 0) ||
 		    (strcmp(blkdev->d_name,"..") == 0))
 			continue;
 
-		strcat(devptr,blkdev->d_name);
-		if (stat(devptr, &statbuf) < 0)
+		devptr = devpath + 11;
+		*devptr = '\0';
+		strncat(devptr, blkdev->d_name, PATH_MAX-12);
+		if (stat(devpath, &statbuf) < 0)
 			continue;
 
 		if (S_ISDIR(statbuf.st_mode) == 0)
@@ -1135,18 +1289,20 @@ snprint_devices (char * buff, int len, struct vectors *vecs)
 		if ((len - fwd - threshold)  <= 0)
 			return len;
 
-		fwd += snprintf(buff + fwd, len - fwd, "    %s", devpath);
-		pp = find_path_by_dev(vecs->pathvec, devpath);
+		fwd += snprintf(buff + fwd, len - fwd, "    %s", devptr);
+		pp = find_path_by_dev(vecs->pathvec, devptr);
 		if (!pp) {
 			r = filter_devnode(conf->blist_devnode,
-					   conf->elist_devnode, devpath);
+					   conf->elist_devnode, devptr);
 			if (r > 0)
 				fwd += snprintf(buff + fwd, len - fwd,
-						" (blacklisted)");
+						" devnode blacklisted, unmonitored");
 			else if (r < 0)
 				fwd += snprintf(buff + fwd, len - fwd,
-						" (whitelisted)");
-		}
+						" devnode whitelisted, unmonitored");
+		} else
+			fwd += snprintf(buff + fwd, len - fwd,
+					" devnode whitelisted, monitored");
 		fwd += snprintf(buff + fwd, len - fwd, "\n");
 	}
 	closedir(blkdir);
@@ -1170,6 +1326,7 @@ print_path (struct path * pp, char * style)
 {
 	char line[MAX_LINE_LEN];
 
+	memset(&line[0], 0, MAX_LINE_LEN);
 	snprint_path(&line[0], MAX_LINE_LEN, style, pp);
 	printf("%s", line);
 }
@@ -1179,6 +1336,7 @@ print_multipath (struct multipath * mpp, char * style)
 {
 	char line[MAX_LINE_LEN];
 
+	memset(&line[0], 0, MAX_LINE_LEN);
 	snprint_multipath(&line[0], MAX_LINE_LEN, style, mpp);
 	printf("%s", line);
 }
@@ -1188,6 +1346,7 @@ print_pathgroup (struct pathgroup * pgp, char * style)
 {
 	char line[MAX_LINE_LEN];
 
+	memset(&line[0], 0, MAX_LINE_LEN);
 	snprint_pathgroup(&line[0], MAX_LINE_LEN, style, pgp);
 	printf("%s", line);
 }
@@ -1197,12 +1356,18 @@ print_map (struct multipath * mpp)
 {
 	if (mpp->size && mpp->params)
 		printf("0 %llu %s %s\n",
-			 mpp->size, DEFAULT_TARGET, mpp->params);
+			 mpp->size, TGT_MPATH, mpp->params);
 	return;
 }
 
 extern void
 print_all_paths (vector pathvec, int banner)
+{
+	print_all_paths_custo(pathvec, banner, PRINT_PATH_LONG);
+}
+
+extern void
+print_all_paths_custo (vector pathvec, int banner, char *fmt)
 {
 	int i;
 	struct path * pp;
@@ -1213,15 +1378,15 @@ print_all_paths (vector pathvec, int banner)
 			fprintf(stdout, "===== no paths =====\n");
 		return;
 	}
-	
+
 	if (banner)
 		fprintf(stdout, "===== paths list =====\n");
 
-	get_path_layout(pathvec);
-	snprint_path_header(line, MAX_LINE_LEN, PRINT_PATH_LONG);
+	get_path_layout(pathvec, 1);
+	snprint_path_header(line, MAX_LINE_LEN, fmt);
 	fprintf(stdout, "%s", line);
 
 	vector_foreach_slot (pathvec, pp, i)
-		print_path(pp, PRINT_PATH_LONG);
+		print_path(pp, fmt);
 }
 
