@@ -5,8 +5,7 @@
  */
 #include <stdio.h>
 
-#include <checkers.h>
-
+#include "checkers.h"
 #include "memory.h"
 #include "vector.h"
 #include "structs.h"
@@ -16,6 +15,7 @@
 #include "alias.h"
 #include "defaults.h"
 #include "devmapper.h"
+#include "prio.h"
 
 pgpolicyfn *pgpolicies[] = {
 	NULL,
@@ -26,6 +26,60 @@ pgpolicyfn *pgpolicies[] = {
 	group_by_node_name
 };
 
+extern int
+select_mode (struct multipath *mp)
+{
+	if (mp->mpe && (mp->mpe->attribute_flags & (1 << ATTR_MODE))) {
+		mp->attribute_flags |= (1 << ATTR_MODE);
+		mp->mode = mp->mpe->mode;
+		condlog(3, "mode = 0%o (LUN setting)", mp->mode);
+	}
+	else if (conf->attribute_flags & (1 << ATTR_MODE)) {
+		mp->attribute_flags |= (1 << ATTR_MODE);
+		mp->mode = conf->mode;
+		condlog(3, "mode = 0%o (config file default)", mp->mode);
+	}
+	else
+		mp->attribute_flags &= ~(1 << ATTR_MODE);
+	return 0;
+}
+
+extern int
+select_uid (struct multipath *mp)
+{
+	if (mp->mpe && (mp->mpe->attribute_flags & (1 << ATTR_UID))) {
+		mp->attribute_flags |= (1 << ATTR_UID);
+		mp->uid = mp->mpe->uid;
+		condlog(3, "uid = %u (LUN setting)", mp->uid);
+	}
+	else if (conf->attribute_flags & (1 << ATTR_UID)) {
+		mp->attribute_flags |= (1 << ATTR_UID);
+		mp->uid = conf->uid;
+		condlog(3, "uid = %u (config file default)", mp->uid);
+	}
+	else
+		mp->attribute_flags &= ~(1 << ATTR_UID);
+	return 0;
+}
+
+extern int
+select_gid (struct multipath *mp)
+{
+	if (mp->mpe && (mp->mpe->attribute_flags & (1 << ATTR_GID))) {
+		mp->attribute_flags |= (1 << ATTR_GID);
+		mp->gid = mp->mpe->gid;
+		condlog(3, "gid = %u (LUN setting)", mp->gid);
+	}
+	else if (conf->attribute_flags & (1 << ATTR_GID)) {
+		mp->attribute_flags |= (1 << ATTR_GID);
+		mp->gid = conf->gid;
+		condlog(3, "gid = %u (config file default)", mp->gid);
+	}
+	else
+		mp->attribute_flags &= ~(1 << ATTR_GID);
+	return 0;
+}
+
 /*
  * selectors :
  * traverse the configuration layers from most specific to most generic
@@ -149,7 +203,13 @@ select_selector (struct multipath * mp)
 			mp->alias, mp->selector);
 		return 0;
 	}
-	mp->selector = conf->selector;
+	if (conf->selector) {
+		mp->selector = conf->selector;
+		condlog(3, "%s: selector = %s (config file default)",
+			mp->alias, mp->selector);
+		return 0;
+	}
+	mp->selector = set_default(DEFAULT_SELECTOR);
 	condlog(3, "%s: selector = %s (internal default)",
 		mp->alias, mp->selector);
 	return 0;
@@ -168,8 +228,7 @@ select_alias (struct multipath * mp)
 		if (mp->alias == NULL){
 			char *alias;
 			if ((alias = MALLOC(WWID_SIZE)) != NULL){
-				if (dm_get_name(mp->wwid, DEFAULT_TARGET,
-						alias) == 1)
+				if (dm_get_name(mp->wwid, alias) == 1)
 					mp->alias = alias;
 				else
 					FREE(alias);
@@ -191,7 +250,13 @@ select_features (struct multipath * mp)
 			mp->alias, mp->features);
 		return 0;
 	}
-	mp->features = conf->features;
+	if (conf->features) {
+		mp->features = conf->features;
+		condlog(3, "%s: features = %s (config file default)",
+			mp->alias, mp->features);
+		return 0;
+	}
+	mp->features = set_default(DEFAULT_FEATURES);
 	condlog(3, "%s: features = %s (internal default)",
 		mp->alias, mp->features);
 	return 0;
@@ -206,7 +271,13 @@ select_hwhandler (struct multipath * mp)
 			mp->alias, mp->hwhandler);
 		return 0;
 	}
-	mp->hwhandler = conf->hwhandler;
+	if (conf->hwhandler) {
+		mp->hwhandler = conf->hwhandler;
+		condlog(3, "%s: hwhandler = %s (config file default)",
+			mp->alias, mp->hwhandler);
+		return 0;
+	}
+	mp->hwhandler = set_default(DEFAULT_HWHANDLER);
 	condlog(3, "%s: hwhandler = %s (internal default)",
 		mp->alias, mp->hwhandler);
 	return 0;
@@ -217,19 +288,19 @@ select_checker(struct path *pp)
 {
 	struct checker * c = &pp->checker;
 
-	if (pp->hwe && pp->hwe->checker) {
-		checker_get(c, pp->hwe->checker);
+	if (pp->hwe && pp->hwe->checker_name) {
+		checker_get(c, pp->hwe->checker_name);
 		condlog(3, "%s: path checker = %s (controller setting)",
 			pp->dev, checker_name(c));
 		return 0;
 	}
-	if (conf->checker) {
-		checker_get(c, conf->checker);
+	if (conf->checker_name) {
+		checker_get(c, conf->checker_name);
 		condlog(3, "%s: path checker = %s (config file default)",
 			pp->dev, checker_name(c));
 		return 0;
 	}
-	checker_get(c, checker_default());
+	checker_get(c, DEFAULT_CHECKER);
 	condlog(3, "%s: path checker = %s (internal default)",
 		pp->dev, checker_name(c));
 	return 0;
@@ -257,28 +328,42 @@ select_getuid (struct path * pp)
 }
 
 extern int
-select_getprio (struct path * pp)
+select_prio (struct path * pp)
 {
-	if (pp->hwe && pp->hwe->getprio) {
-		pp->getprio = pp->hwe->getprio;
-		condlog(3, "%s: getprio = %s (controller setting)",
-			pp->dev, pp->getprio);
+	if (pp->hwe && pp->hwe->prio_name) {
+		pp->prio = prio_lookup(pp->hwe->prio_name);
+		prio_set_args(pp->prio, pp->hwe->prio_args);
+		condlog(3, "%s: prio = %s (controller setting)",
+			pp->dev, pp->hwe->prio_name);
+		condlog(3, "%s: prio args = %s (controller setting)",
+			pp->dev, pp->hwe->prio_args);
 		return 0;
 	}
-	if (conf->getprio) {
-		pp->getprio = conf->getprio;
-		condlog(3, "%s: getprio = %s (config file default)",
-			pp->dev, pp->getprio);
+	if (conf->prio_name) {
+		pp->prio = prio_lookup(conf->prio_name);
+		prio_set_args(pp->prio, conf->prio_args);
+		condlog(3, "%s: prio = %s (config file default)",
+			pp->dev, conf->prio_name);
+		condlog(3, "%s: prio args = %s (config file default)",
+			pp->dev, conf->prio_args);
 		return 0;
 	}
-	pp->getprio = DEFAULT_GETPRIO;
-	condlog(3, "%s: getprio = NULL (internal default)", pp->dev);
+	pp->prio = prio_lookup(DEFAULT_PRIO);
+        prio_set_args(pp->prio, DEFAULT_PRIO_ARGS);
+	condlog(3, "%s: prio = %s (internal default)",
+		pp->dev, DEFAULT_PRIO);
+	condlog(3, "%s: prio = %s (internal default)",
+		pp->dev, DEFAULT_PRIO_ARGS);
 	return 0;
 }
 
 extern int
 select_no_path_retry(struct multipath *mp)
 {
+	if (mp->flush_on_last_del == FLUSH_IN_PROGRESS) {
+		condlog(0, "flush_on_last_del in progress");
+		mp->no_path_retry = NO_PATH_RETRY_FAIL;
+	}
 	if (mp->mpe && mp->mpe->no_path_retry != NO_PATH_RETRY_UNDEF) {
 		mp->no_path_retry = mp->mpe->no_path_retry;
 		condlog(3, "%s: no_path_retry = %i (multipath setting)",
@@ -368,3 +453,73 @@ select_pg_timeout(struct multipath *mp)
 	condlog(3, "pg_timeout = NONE (internal default)");
 	return 0;
 }
+
+extern int
+select_fast_io_fail(struct multipath *mp)
+{
+	if (mp->hwe && mp->hwe->fast_io_fail) {
+		mp->fast_io_fail = mp->hwe->fast_io_fail;
+		if (mp->fast_io_fail == -1)
+			condlog(3, "%s: fast_io_fail_tmo = off (controller default)", mp->alias);
+		else
+			condlog(3, "%s: fast_io_fail_tmo = %d (controller default)", mp->alias, mp->fast_io_fail);
+		return 0;
+	}
+	if (conf->fast_io_fail) {
+		mp->fast_io_fail = conf->fast_io_fail;
+		if (mp->fast_io_fail == -1)
+			condlog(3, "%s: fast_io_fail_tmo = off (config file default)", mp->alias);
+		else
+			condlog(3, "%s: fast_io_fail_tmo = %d (config file default)", mp->alias, mp->fast_io_fail);
+		return 0;
+	}
+	mp->fast_io_fail = 0;
+	return 0;
+}
+
+extern int
+select_dev_loss(struct multipath *mp)
+{
+	if (mp->hwe && mp->hwe->dev_loss) {
+		mp->dev_loss = mp->hwe->dev_loss;
+		condlog(3, "%s: dev_loss_tmo = %u (controller default)",
+			mp->alias, mp->dev_loss);
+		return 0;
+	}
+	if (conf->dev_loss) {
+		mp->dev_loss = conf->dev_loss;
+		condlog(3, "%s: dev_loss_tmo = %u (config file default)",
+			mp->alias, mp->dev_loss);
+		return 0;
+	}
+	mp->dev_loss = 0;
+	return 0;
+}
+
+extern int
+select_flush_on_last_del(struct multipath *mp)
+{
+	if (mp->flush_on_last_del == FLUSH_IN_PROGRESS)
+		return 0;
+	if (mp->mpe && mp->mpe->flush_on_last_del != FLUSH_UNDEF) {
+		mp->flush_on_last_del = mp->mpe->flush_on_last_del;
+		condlog(3, "flush_on_last_del = %i (multipath setting)",
+				mp->flush_on_last_del);
+		return 0;
+	}
+	if (mp->hwe && mp->hwe->flush_on_last_del != FLUSH_UNDEF) {
+		mp->flush_on_last_del = mp->hwe->flush_on_last_del;
+		condlog(3, "flush_on_last_del = %i (controler setting)",
+				mp->flush_on_last_del);
+		return 0;
+	}
+	if (conf->flush_on_last_del != FLUSH_UNDEF) {
+		mp->flush_on_last_del = conf->flush_on_last_del;
+		condlog(3, "flush_on_last_del = %i (config file default)",
+				mp->flush_on_last_del);
+		return 0;
+	}
+	mp->flush_on_last_del = FLUSH_UNDEF;
+	condlog(3, "flush_on_last_del = DISABLED (internal default)");
+	return 0;
+}
