@@ -6,8 +6,7 @@
 #include <unistd.h>
 #include <libdevmapper.h>
 
-#include <checkers.h>
-
+#include "checkers.h"
 #include "memory.h"
 #include "vector.h"
 #include "util.h"
@@ -17,12 +16,13 @@
 #include "structs_vec.h"
 #include "blacklist.h"
 #include "waiter.h"
+#include "prio.h"
 
 struct path *
 alloc_path (void)
 {
 	struct path * pp;
-	
+
 	pp = (struct path *)MALLOC(sizeof(struct path));
 
 	if (pp) {
@@ -79,8 +79,10 @@ alloc_pathgroup (void)
 
 	pgp->paths = vector_alloc();
 
-	if (!pgp->paths)
+	if (!pgp->paths) {
 		FREE(pgp);
+		pgp = NULL;
+	}
 
 	return pgp;
 }
@@ -168,16 +170,16 @@ free_multipath (struct multipath * mpp, int free_paths)
 		mpp->alias = NULL;
 	}
 
-	if (mpp->dmi)
+	if (mpp->dmi) {
 		FREE(mpp->dmi);
-	
-#if DAEMON
+		mpp->dmi = NULL;
+	}
+
 	/*
 	 * better own vecs->lock here
 	 */
 	if (mpp->waiter)
 		((struct event_thread *)mpp->waiter)->mpp = NULL;
-#endif
 
 	free_pathvec(mpp->paths, free_paths);
 	free_pgvec(mpp->pg, free_paths);
@@ -245,7 +247,7 @@ find_mp_by_minor (vector mpvec, int minor)
 {
 	int i;
 	struct multipath * mpp;
-	
+
 	if (!mpvec)
 		return NULL;
 
@@ -264,7 +266,7 @@ find_mp_by_wwid (vector mpvec, char * wwid)
 {
 	int i;
 	struct multipath * mpp;
-	
+
 	if (!mpvec)
 		return NULL;
 
@@ -281,7 +283,7 @@ find_mp_by_alias (vector mpvec, char * alias)
 	int i;
 	int len;
 	struct multipath * mpp;
-	
+
 	if (!mpvec)
 		return NULL;
 
@@ -289,7 +291,7 @@ find_mp_by_alias (vector mpvec, char * alias)
 
 	if (!len)
 		return NULL;
-	
+
 	vector_foreach_slot (mpvec, mpp, i) {
 		if (strlen(mpp->alias) == len &&
 		    !strncmp(mpp->alias, alias, len))
@@ -317,7 +319,7 @@ find_path_by_dev (vector pathvec, char * dev)
 
 	if (!pathvec)
 		return NULL;
-	
+
 	vector_foreach_slot (pathvec, pp, i)
 		if (!strcmp_chomp(pp->dev, dev))
 			return pp;
@@ -351,7 +353,7 @@ pathcountgr (struct pathgroup * pgp, int state)
 	int i;
 
 	vector_foreach_slot (pgp->paths, pp, i)
-		if ((pp->state == state) || (state < 0))
+		if ((pp->state == state) || (state == PATH_WILD))
 			count++;
 
 	return count;
@@ -364,15 +366,27 @@ pathcount (struct multipath * mpp, int state)
 	int count = 0;
 	int i;
 
-	vector_foreach_slot (mpp->pg, pgp, i)
-		count += pathcountgr(pgp, state);
-
+	if (mpp->pg) {
+		vector_foreach_slot (mpp->pg, pgp, i)
+			count += pathcountgr(pgp, state);
+	}
 	return count;
 }
 
 struct path *
 first_path (struct multipath * mpp)
 {
-	struct pathgroup * pgp = VECTOR_SLOT(mpp->pg, 0);
-	return VECTOR_SLOT(pgp->paths, 0);
+	struct pathgroup * pgp;
+	if (!mpp->pg)
+		return NULL;
+	pgp = VECTOR_SLOT(mpp->pg, 0);
+
+	return pgp?VECTOR_SLOT(pgp->paths, 0):NULL;
+}
+
+extern void
+setup_feature(struct multipath * mpp, char *feature)
+{
+	if (!strncmp(feature, "queue_if_no_path", 16))
+		mpp->no_path_retry = NO_PATH_RETRY_QUEUE;
 }
