@@ -2,20 +2,20 @@
 #include <string.h>
 #include <unistd.h>
 
-#include <checkers.h>
-
+#include "checkers.h"
 #include "vector.h"
 #include "defaults.h"
 #include "debug.h"
 #include "structs.h"
 #include "structs_vec.h"
+#include "waiter.h"
 #include "devmapper.h"
 #include "dmparser.h"
 #include "config.h"
 #include "propsel.h"
 #include "sysfs.h"
 #include "discovery.h"
-#include "waiter.h"
+#include "prio.h"
 
 /*
  * creates or updates mpp->paths reading mpp->pg
@@ -27,7 +27,7 @@ update_mpp_paths(struct multipath * mpp, vector pathvec)
 	struct path * pp;
 	int i,j;
 
-	if (!mpp->pg)
+	if (!mpp || !mpp->pg)
 		return 0;
 
 	if (!mpp->paths &&
@@ -62,7 +62,7 @@ adopt_paths (vector pathvec, struct multipath * mpp)
 			condlog(3, "%s: ownership set to %s",
 				pp->dev, mpp->alias);
 			pp->mpp = mpp;
-			
+
 			if (!mpp->paths && !(mpp->paths = vector_alloc()))
 				return 1;
 
@@ -81,8 +81,7 @@ orphan_path (struct path * pp)
 	pp->mpp = NULL;
 	pp->dmstate = PSTATE_UNDEF;
 	pp->getuid = NULL;
-	pp->getprio = NULL;
-	pp->getprio_selected = 0;
+	pp->prio = NULL;
 	checker_put(&pp->checker);
 	if (pp->fd >= 0)
 		close(pp->fd);
@@ -112,9 +111,13 @@ set_multipath_wwid (struct multipath * mpp)
 	dm_get_uuid(mpp->alias, mpp->wwid);
 }
 
-extern void
-remove_map (struct multipath * mpp, struct vectors * vecs,
-	    stop_waiter_thread_func *stop_waiter, int purge_vec)
+#define KEEP_WAITER 0
+#define STOP_WAITER 1
+#define PURGE_VEC 1
+
+static void
+_remove_map (struct multipath * mpp, struct vectors * vecs,
+	    int stop_waiter, int purge_vec)
 {
 	int i;
 
@@ -124,7 +127,7 @@ remove_map (struct multipath * mpp, struct vectors * vecs,
 	 * stop the DM event waiter thread
 	 */
 	if (stop_waiter)
-		stop_waiter(mpp, vecs);
+		stop_waiter_thread(mpp, vecs);
 
 	/*
 	 * clear references to this map
@@ -142,14 +145,29 @@ remove_map (struct multipath * mpp, struct vectors * vecs,
 }
 
 extern void
-remove_maps (struct vectors * vecs,
-	     stop_waiter_thread_func *stop_waiter)
+remove_map (struct multipath * mpp, struct vectors * vecs, int purge_vec)
+{
+	_remove_map(mpp, vecs, KEEP_WAITER, purge_vec);
+}
+
+extern void
+remove_map_and_stop_waiter (struct multipath * mpp, struct vectors * vecs,
+			    int purge_vec)
+{
+	_remove_map(mpp, vecs, STOP_WAITER, purge_vec);
+}
+
+static void
+_remove_maps (struct vectors * vecs, int stop_waiter)
 {
 	int i;
 	struct multipath * mpp;
 
+	if (!vecs)
+		return;
+
 	vector_foreach_slot (vecs->mpvec, mpp, i) {
-		remove_map(mpp, vecs, stop_waiter, 1);
+		_remove_map(mpp, vecs, stop_waiter, 1);
 		i--;
 	}
 
@@ -157,16 +175,66 @@ remove_maps (struct vectors * vecs,
 	vecs->mpvec = NULL;
 }
 
+extern void
+remove_maps (struct vectors * vecs)
+{
+	_remove_maps(vecs, KEEP_WAITER);
+}
+
+extern void
+remove_maps_and_stop_waiters (struct vectors * vecs)
+{
+	_remove_maps(vecs, STOP_WAITER);
+}
+
 static struct hwentry *
 extract_hwe_from_path(struct multipath * mpp)
 {
-	struct path * pp;
-	struct pathgroup * pgp;
+	struct path * pp = NULL;
+	int pg_num = -1, p_num = -1, i;
+	struct pathgroup * pgp = NULL;
 
-	pgp = VECTOR_SLOT(mpp->pg, 0);
-	pp = VECTOR_SLOT(pgp->paths, 0);
+	condlog(3, "%s: searching paths for valid hwe", mpp->alias);
 
-	return pp->hwe;
+	if (mpp && mpp->pg) {
+		vector_foreach_slot(mpp->pg, pgp, i) {
+			if (pgp->status == PGSTATE_ACTIVE ||
+			    pgp->status == PGSTATE_ENABLED) {
+				pg_num = i;
+				break;
+			}
+		}
+		if (pg_num >= 0)
+			pgp = VECTOR_SLOT(mpp->pg, pg_num);
+	}
+
+	if (pgp && pgp->paths) {
+		vector_foreach_slot(pgp->paths, pp, i) {
+			if (pp->dmstate == PSTATE_FAILED)
+				continue;
+			if (strlen(pp->vendor_id) > 0 &&
+			    strlen(pp->product_id) > 0 &&
+			    strlen(pp->rev) > 0) {
+				p_num = i;
+				break;
+			}
+		}
+		if (p_num >= 0)
+			pp = VECTOR_SLOT(pgp->paths, i);
+	}
+
+	if (pp) {
+		condlog(3, "%s: vendor = %s", pp->dev, pp->vendor_id);
+		condlog(3, "%s: product = %s", pp->dev, pp->product_id);
+		condlog(3, "%s: rev = %s", pp->dev, pp->rev);
+		if (!pp->hwe) {
+			condlog(3, "searching hwtable");
+			pp->hwe = find_hwe(conf->hwtable, pp->vendor_id,
+					   pp->product_id, pp->rev);
+		}
+	}
+
+	return pp?pp->hwe:NULL;
 }
 
 static int
@@ -202,6 +270,8 @@ update_multipath_status (struct multipath *mpp)
 extern int
 update_multipath_strings (struct multipath *mpp, vector pathvec)
 {
+	condlog(4, "%s: %s", mpp->alias, __FUNCTION__);
+
 	free_multipath_attributes(mpp);
 	free_pgvec(mpp->pg, KEEP_PATHS);
 	mpp->pg = NULL;
@@ -220,7 +290,8 @@ set_no_path_retry(struct multipath *mpp)
 {
 	mpp->retry_tick = 0;
 	mpp->nr_active = pathcount(mpp, PATH_UP) + pathcount(mpp, PATH_GHOST);
-	select_no_path_retry(mpp);
+	if (mpp->nr_active > 0)
+		select_no_path_retry(mpp);
 
 	switch (mpp->no_path_retry) {
 	case NO_PATH_RETRY_UNDEF:
@@ -249,13 +320,13 @@ setup_multipath (struct vectors * vecs, struct multipath * mpp)
 retry:
 	if (dm_get_info(mpp->alias, &mpp->dmi)) {
 		/* Error accessing table */
-		condlog(3, "%s: cannot access table", mpp->alias); 
+		condlog(3, "%s: cannot access table", mpp->alias);
 		goto out;
 	}
 
 	if (!dm_map_present(mpp->alias)) {
 		/* Table has been removed */
-		condlog(3, "%s: table does not exist", mpp->alias); 
+		condlog(3, "%s: table does not exist", mpp->alias);
 		goto out;
 	}
 
@@ -267,18 +338,17 @@ retry:
 		char new_alias[WWID_SIZE];
 
 		/*
-	 	 * detect an external rename of the multipath device
+		 * detect an external rename of the multipath device
 		 */
-		if (dm_get_name(mpp->wwid, DEFAULT_TARGET, new_alias)) {
+		if (dm_get_name(mpp->wwid, new_alias)) {
 			condlog(3, "%s multipath mapped device name has "
 				"changed from %s to %s", mpp->wwid,
 				mpp->alias, new_alias);
 			strcpy(mpp->alias, new_alias);
-#if DAEMON
-			if (mpp->waiter) 
+
+			if (mpp->waiter)
 				strncpy(((struct event_thread *)mpp->waiter)->mapname,
 					new_alias, WWID_SIZE);
-#endif
 			goto retry;
 		}
 		condlog(0, "%s: failed to setup multipath", mpp->alias);
@@ -286,22 +356,27 @@ retry:
 	}
 
 	//adopt_paths(vecs->pathvec, mpp);
-	mpp->hwe = extract_hwe_from_path(mpp);
+	if (!mpp->hwe)
+		mpp->hwe = extract_hwe_from_path(mpp);
+	if (!mpp->hwe) {
+		condlog(3, "%s: no hardware entry found, using defaults",
+			mpp->alias);
+	}
 	select_rr_weight(mpp);
 	select_pgfailback(mpp);
 	set_no_path_retry(mpp);
 	select_pg_timeout(mpp);
+	select_flush_on_last_del(mpp);
 
 	return 0;
 out:
-	remove_map(mpp, vecs, NULL, 1);
+	remove_map(mpp, vecs, PURGE_VEC);
 	return 1;
 }
 
 extern struct multipath *
 add_map_without_path (struct vectors * vecs,
-		      int minor, char * alias,
-		      start_waiter_thread_func *start_waiter)
+		      int minor, char * alias)
 {
 	struct multipath * mpp = alloc_multipath();
 
@@ -310,23 +385,25 @@ add_map_without_path (struct vectors * vecs,
 
 	mpp->alias = alias;
 
-	if (setup_multipath(vecs, mpp))
+	if (setup_multipath(vecs, mpp)) {
+		mpp->alias = NULL;
 		return NULL; /* mpp freed in setup_multipath */
+	}
 
 	if (adopt_paths(vecs->pathvec, mpp))
 		goto out;
-	
+
 	if (!vector_alloc_slot(vecs->mpvec))
 		goto out;
 
 	vector_set_slot(vecs->mpvec, mpp);
 
-	if (start_waiter(mpp, vecs))
+	if (start_waiter_thread(mpp, vecs))
 		goto out;
 
 	return mpp;
 out:
-	remove_map(mpp, vecs, NULL, 1);
+	remove_map(mpp, vecs, PURGE_VEC);
 	return NULL;
 }
 
@@ -359,7 +436,7 @@ add_map_with_path (struct vectors * vecs,
 	return mpp;
 
 out:
-	remove_map(mpp, vecs, NULL, add_vec);
+	remove_map(mpp, vecs, PURGE_VEC);
 	return NULL;
 }
 
@@ -370,6 +447,9 @@ verify_paths(struct multipath * mpp, struct vectors * vecs, vector rpvec)
 	int count = 0;
 	int i, j;
 
+	if (!mpp)
+		return 0;
+
 	vector_foreach_slot (mpp->paths, pp, i) {
 		/*
 		 * see if path is in sysfs
@@ -390,6 +470,9 @@ verify_paths(struct multipath * mpp, struct vectors * vecs, vector rpvec)
 					vector_del_slot(vecs->pathvec, j);
 				free_path(pp);
 			}
+		} else {
+			condlog(4, "%s: verified path %s dev_t %s",
+				mpp->alias, pp->dev, pp->dev_t);
 		}
 	}
 	return count;
