@@ -6,7 +6,7 @@
 #define _LIBMULTIPATH_SYSFS_H
 
 #ifdef DEBUG
-# define dbg(format, args...) printf(format "\n", ##args)
+# define dbg(format, args...) condlog(4, format, ##args)
 #else
 # define dbg(format, args...) do {} while (0)
 #endif
@@ -21,5 +21,7 @@ struct sysfs_device *sysfs_device_get_parent_with_subsystem(struct sysfs_device
 void sysfs_device_put(struct sysfs_device *dev);
 char *sysfs_attr_get_value(const char *devpath, const char *attr_name);
 int sysfs_resolve_link(char *path, size_t size);
-
+int sysfs_get_size (struct sysfs_device * dev, unsigned long long * size);
+int sysfs_attr_set_value(const char *devpath, const char *attr_name,
+			 const char *value);
 #endif
