@@ -25,6 +25,7 @@
 
 #include <unistd.h>
 #include <stdio.h>
+#include <errno.h>
 #include <stdlib.h>
 #include <stddef.h>
 #include <string.h>
@@ -51,8 +52,14 @@ pthread_mutex_t uevc_lock, *uevc_lockp = &uevc_lock;
 pthread_cond_t  uev_cond,  *uev_condp  = &uev_cond;
 uev_trigger *my_uev_trigger;
 void * my_trigger_data;
+int servicing_uev;
 
-struct uevent * alloc_uevent (void)
+int is_uevent_busy(void)
+{
+	return (uevqhp != NULL || servicing_uev);
+}
+
+static struct uevent * alloc_uevent (void)
 {
 	return (struct uevent *)MALLOC(sizeof(struct uevent));
 }
@@ -95,11 +102,14 @@ uevq_thread(void * et)
 
 	while (1) {
 		pthread_mutex_lock(uevc_lockp);
+		servicing_uev = 0;
 		pthread_cond_wait(uev_condp, uevc_lockp);
+		servicing_uev = 1;
 		pthread_mutex_unlock(uevc_lockp);
 
 		service_uevq();
 	}
+	return NULL;
 }
 
 int uevent_listen(int (*uev_trigger)(struct uevent *, void * trigger_data),
@@ -115,6 +125,7 @@ int uevent_listen(int (*uev_trigger)(struct uevent *, void * trigger_data),
 	int rcvszsz = sizeof(rcvsz);
 	unsigned int *prcvszsz = (unsigned int *)&rcvszsz;
 	pthread_attr_t attr;
+	const int feature_on = 1;
 
 	my_uev_trigger = uev_trigger;
 	my_trigger_data = trigger_data;
@@ -145,7 +156,6 @@ int uevent_listen(int (*uev_trigger)(struct uevent *, void * trigger_data),
 
 	sock = socket(AF_LOCAL, SOCK_DGRAM, 0);
 	if (sock >= 0) {
-		const int feature_on = 1;
 
 		condlog(3, "reading events from udev socket.");
 
@@ -157,7 +167,7 @@ int uevent_listen(int (*uev_trigger)(struct uevent *, void * trigger_data),
 		}
 
 		/* enable receiving of the sender credentials */
-		setsockopt(sock, SOL_SOCKET, SO_PASSCRED, 
+		setsockopt(sock, SOL_SOCKET, SO_PASSCRED,
 			   &feature_on, sizeof(feature_on));
 
 	} else {
@@ -165,7 +175,7 @@ int uevent_listen(int (*uev_trigger)(struct uevent *, void * trigger_data),
 		memset(&snl, 0x00, sizeof(struct sockaddr_nl));
 		snl.nl_family = AF_NETLINK;
 		snl.nl_pid = getpid();
-		snl.nl_groups = 0xffffffff;
+		snl.nl_groups = 0x01;
 
 		sock = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_KOBJECT_UEVENT);
 		if (sock == -1) {
@@ -195,6 +205,10 @@ int uevent_listen(int (*uev_trigger)(struct uevent *, void * trigger_data),
 		}
 		condlog(3, "receive buffer size for socket is %u.", rcvsz);
 
+		/* enable receiving of the sender credentials */
+		setsockopt(sock, SOL_SOCKET, SO_PASSCRED,
+			   &feature_on, sizeof(feature_on));
+
 		retval = bind(sock, (struct sockaddr *) &snl,
 			      sizeof(struct sockaddr_nl));
 		if (retval < 0) {
@@ -204,22 +218,59 @@ int uevent_listen(int (*uev_trigger)(struct uevent *, void * trigger_data),
 	}
 
 	while (1) {
-		static char buff[HOTPLUG_BUFFER_SIZE + OBJECT_SIZE];
 		int i;
 		char *pos;
 		size_t bufpos;
 		ssize_t buflen;
 		struct uevent *uev;
 		char *buffer;
+		struct msghdr smsg;
+		struct iovec iov;
+		char cred_msg[CMSG_SPACE(sizeof(struct ucred))];
+		struct cmsghdr *cmsg;
+		struct ucred *cred;
+		static char buf[HOTPLUG_BUFFER_SIZE + OBJECT_SIZE];
 
-		buflen = recv(sock, &buff, sizeof(buff), 0);
-		if (buflen <  0) {
-			condlog(0, "error receiving message");
+		memset(buf, 0x00, sizeof(buf));
+		iov.iov_base = &buf;
+		iov.iov_len = sizeof(buf);
+		memset (&smsg, 0x00, sizeof(struct msghdr));
+		smsg.msg_iov = &iov;
+		smsg.msg_iovlen = 1;
+		smsg.msg_control = cred_msg;
+		smsg.msg_controllen = sizeof(cred_msg);
+
+		buflen = recvmsg(sock, &smsg, 0);
+		if (buflen < 0) {
+			if (errno != EINTR)
+				condlog(0, "error receiving message");
 			continue;
 		}
 
-		if ((size_t)buflen > sizeof(buff)-1)
-			buflen = sizeof(buff)-1;
+		cmsg = CMSG_FIRSTHDR(&smsg);
+		if (cmsg == NULL || cmsg->cmsg_type != SCM_CREDENTIALS) {
+			condlog(3, "no sender credentials received, message ignored");
+			continue;
+		}
+
+		cred = (struct ucred *)CMSG_DATA(cmsg);
+		if (cred->uid != 0) {
+			condlog(3, "sender uid=%d, message ignored", cred->uid);
+			continue;
+		}
+
+		/* skip header */
+		bufpos = strlen(buf) + 1;
+		if (bufpos < sizeof("a@/d") || bufpos >= sizeof(buf)) {
+			condlog(3, "invalid message length");
+			continue;
+		}
+
+		/* check message header */
+		if (strstr(buf, "@/") == NULL) {
+			condlog(3, "unrecognized message header");
+			continue;
+		}
 
 		uev = alloc_uevent();
 
@@ -228,11 +279,14 @@ int uevent_listen(int (*uev_trigger)(struct uevent *, void * trigger_data),
 			continue;
 		}
 
+		if ((size_t)buflen > sizeof(buf)-1)
+			buflen = sizeof(buf)-1;
+
 		/*
 		 * Copy the shared receive buffer contents to buffer private
 		 * to this uevent so we can immediately reuse the shared buffer.
 		 */
-		memcpy(uev->buffer, buff, HOTPLUG_BUFFER_SIZE + OBJECT_SIZE);
+		memcpy(uev->buffer, buf, HOTPLUG_BUFFER_SIZE + OBJECT_SIZE);
 		buffer = uev->buffer;
 		buffer[buflen] = '\0';
 
@@ -242,8 +296,10 @@ int uevent_listen(int (*uev_trigger)(struct uevent *, void * trigger_data),
 		/* action string */
 		uev->action = buffer;
 		pos = strchr(buffer, '@');
-		if (!pos)
+		if (!pos) {
+			condlog(3, "bad action string '%s'", buffer);
 			continue;
+		}
 		pos[0] = '\0';
 
 		/* sysfs path */
