@@ -14,7 +14,7 @@ strcmp_chomp(char *str1, char *str2)
 {
 	int i;
 	char s1[PARAMS_SIZE],s2[PARAMS_SIZE];
- 
+
 	if(!str1 || !str2)
 		return 1;
 
@@ -39,7 +39,7 @@ strchop(char *str)
 }
 
 void
-basename (char * str1, char * str2)
+basenamecpy (char * str1, char * str2)
 {
 	char *p = str1 + (strlen(str1) - 1);
 
@@ -67,7 +67,7 @@ get_word (char * sentence, char ** word)
 	char * p;
 	int len;
 	int skip = 0;
-	
+
 	if (word)
 		*word = NULL;
 
@@ -95,6 +95,7 @@ get_word (char * sentence, char ** word)
 		return 0;
 	}
 	strncpy(*word, sentence, len);
+	strchop(*word);
 	condlog(4, "*word = %s, len = %i", *word, len);
 
 	if (*p == '\0')
