@@ -27,6 +27,7 @@
 #include <ctype.h>
 
 #include <checkers.h>
+#include <prio.h>
 #include <vector.h>
 #include <memory.h>
 #include <libdevmapper.h>
@@ -47,6 +48,11 @@
 #include <configure.h>
 #include <pgpolicies.h>
 #include <version.h>
+#include <errno.h>
+#include <sys/time.h>
+#include <sys/resource.h>
+
+int logsink;
 
 static int
 filter_pathvec (vector pathvec, char * refwwid)
@@ -72,34 +78,39 @@ static void
 usage (char * progname)
 {
 	fprintf (stderr, VERSION_STRING);
-	fprintf (stderr, "Usage: %s\t[-v level] [-d] [-h|-l|-ll|-f|-F]\n",
-		progname);
+	fprintf (stderr, "Usage:\n");
+	fprintf (stderr, "  %s [-d] [-r] [-v lvl] [-p pol] [-b fil] [dev]\n", progname);
+	fprintf (stderr, "  %s -l|-ll|-f [-v lvl] [-b fil] [dev]\n", progname);
+	fprintf (stderr, "  %s -F [-v lvl]\n", progname);
+	fprintf (stderr, "  %s -h\n", progname);
 	fprintf (stderr,
-		"\t\t\t[-p failover|multibus|group_by_serial|group_by_prio]\n" \
-		"\t\t\t[device]\n" \
-		"\n" \
-		"\t-v level\tverbosity level\n" \
-		"\t   0\t\t\tno output\n" \
-		"\t   1\t\t\tprint created devmap names only\n" \
-		"\t   2\t\t\tdefault verbosity\n" \
-		"\t   3\t\t\tprint debug information\n" \
-		"\t-h\t\tprint this usage text\n" \
-		"\t-b file\t\tbindings file location\n" \
-		"\t-d\t\tdry run, do not create or update devmaps\n" \
-		"\t-l\t\tshow multipath topology (sysfs and DM info)\n" \
-		"\t-ll\t\tshow multipath topology (maximum info)\n" \
-		"\t-f\t\tflush a multipath device map\n" \
-		"\t-F\t\tflush all multipath device maps\n" \
-		"\t-p policy\tforce all maps to specified policy :\n" \
-		"\t   failover\t\t1 path per priority group\n" \
-		"\t   multibus\t\tall paths in 1 priority group\n" \
-		"\t   group_by_serial\t1 priority group per serial\n" \
-		"\t   group_by_prio\t1 priority group per priority lvl\n" \
-		"\t   group_by_node_name\t1 priority group per target node\n" \
-		"\n" \
-		"\tdevice\t\tlimit scope to the device's multipath\n" \
-		"\t\t\t(udev-style $DEVNAME reference, eg /dev/sdb\n" \
-		"\t\t\tor major:minor or a device map name)\n" \
+		"\n"
+		"Where:\n"
+		"  -h      print this usage text\n" \
+		"  -l      show multipath topology (sysfs and DM info)\n" \
+		"  -ll     show multipath topology (maximum info)\n" \
+		"  -f      flush a multipath device map\n" \
+		"  -F      flush all multipath device maps\n" \
+		"  -d      dry run, do not create or update devmaps\n" \
+		"  -r      force devmap reload\n" \
+		"  -p      policy failover|multibus|group_by_serial|group_by_prio\n" \
+		"  -b fil  bindings file location\n" \
+		"  -p pol  force all maps to specified path grouping policy :\n" \
+		"          . failover            one path per priority group\n" \
+		"          . multibus            all paths in one priority group\n" \
+		"          . group_by_serial     one priority group per serial\n" \
+		"          . group_by_prio       one priority group per priority lvl\n" \
+		"          . group_by_node_name  one priority group per target node\n" \
+		"  -v lvl  verbosity level\n" \
+		"          . 0 no output\n" \
+		"          . 1 print created devmap names only\n" \
+		"          . 2 default verbosity\n" \
+		"          . 3 print debug information\n" \
+		"  dev     action limited to:\n" \
+		"          . multipath named 'dev' (ex: mpath0) or\n" \
+		"          . multipath whose wwid is 'dev' (ex: 60051..)\n" \
+		"          . multipath including the path named 'dev' (ex: /dev/sda)\n" \
+		"          . multipath including the path with maj:min 'dev' (ex: 8:0)\n" \
 		);
 
 	exit(1);
@@ -133,7 +144,8 @@ update_paths (struct multipath * mpp)
 				continue;
 			}
 			pp->mpp = mpp;
-			if (pp->state == PATH_UNCHECKED)
+			if (pp->state == PATH_UNCHECKED ||
+			    pp->state == PATH_WILD)
 				pathinfo(pp, conf->hwtable, DI_CHECKER);
 
 			if (pp->priority == PRIO_UNDEF)
@@ -149,7 +161,7 @@ get_dm_mpvec (vector curmp, vector pathvec, char * refwwid)
 	int i;
 	struct multipath * mpp;
 
-	if (dm_get_maps(curmp, DEFAULT_TARGET))
+	if (dm_get_maps(curmp))
 		return 1;
 
 	vector_foreach_slot (curmp, mpp, i) {
@@ -280,7 +292,7 @@ configure (void)
 	if (conf->verbosity > 2)
 		print_all_paths(pathvec, 1);
 
-	get_path_layout(pathvec);
+	get_path_layout(pathvec, 0);
 
 	if (get_dm_mpvec(curmp, pathvec, refwwid))
 		goto out;
@@ -295,7 +307,7 @@ configure (void)
 	/*
 	 * core logic entry point
 	 */
-	r = coalesce_paths(&vecs, NULL, NULL);
+	r = coalesce_paths(&vecs, NULL, NULL, conf->force_reload);
 
 out:
 	if (refwwid)
@@ -320,17 +332,25 @@ main (int argc, char *argv[])
 		exit(1);
 	}
 
-	if (dm_prereq(DEFAULT_TARGET))
+	if (dm_prereq())
 		exit(1);
 
 	if (load_config(DEFAULT_CONFIGFILE))
 		exit(1);
 
+	if (init_checkers()) {
+		condlog(0, "failed to initialize checkers");
+		exit(1);
+	}
+	if (init_prio()) {
+		condlog(0, "failed to initialize prioritizers");
+		exit(1);
+	}
 	if (sysfs_init(conf->sysfs_dir, FILE_NAME_SIZE)) {
 		condlog(0, "multipath tools need sysfs mounted");
 		exit(1);
 	}
-	while ((arg = getopt(argc, argv, ":dhl::FfM:v:p:b:")) != EOF ) {
+	while ((arg = getopt(argc, argv, ":dhl::FfM:v:p:b:r")) != EOF ) {
 		switch(arg) {
 		case 1: printf("optarg : %s\n",optarg);
 			break;
@@ -373,6 +393,9 @@ main (int argc, char *argv[])
 				usage(argv[0]);
 			}                
 			break;
+		case 'r':
+			conf->force_reload = 1;
+			break;
 		case 'h':
 			usage(argv[0]);
 		case ':':
@@ -401,28 +424,52 @@ main (int argc, char *argv[])
 			conf->dev_type = DEV_DEVMAP;
 
 	}
+	conf->daemon = 0;
+
+	if (conf->max_fds) {
+		struct rlimit fd_limit;
+
+		fd_limit.rlim_cur = conf->max_fds;
+		fd_limit.rlim_max = conf->max_fds;
+		if (setrlimit(RLIMIT_NOFILE, &fd_limit) < 0)
+			condlog(0, "can't set open fds limit to %d : %s\n",
+				conf->max_fds, strerror(errno));
+	}
+
 	dm_init();
 
 	if (conf->remove == FLUSH_ONE) {
 		if (conf->dev_type == DEV_DEVMAP)
-			dm_flush_map(conf->dev, DEFAULT_TARGET);
+			r = dm_flush_map(conf->dev);
 		else
 			condlog(0, "must provide a map name to remove");
 
 		goto out;
 	}
 	else if (conf->remove == FLUSH_ALL) {
-		dm_flush_maps(DEFAULT_TARGET);
+		r = dm_flush_maps();
 		goto out;
 	}
 	while ((r = configure()) < 0)
 		condlog(3, "restart multipath configuration process");
 	
 out:
+	dm_udev_wait(conf->cookie);
+
 	sysfs_cleanup();
-	free_config(conf);
 	dm_lib_release();
 	dm_lib_exit();
+
+	cleanup_prio();
+	cleanup_checkers();
+	/*
+	 * Freeing config must be done after dm_lib_exit(), because
+	 * the logging function (dm_write_log()), which is called there,
+	 * references the config.
+	 */
+	free_config(conf);
+	conf = NULL;
+
 #ifdef _DEBUG_
 	dbg_free_final(NULL);
 #endif
