@@ -3,6 +3,7 @@
  */
 #include <memory.h>
 #include <vector.h>
+#include <parser.h>
 #include <util.h>
 #include <version.h>
 #include <readline/readline.h>
@@ -154,6 +155,9 @@ load_keys (void)
 	r += add_key(keys, "resume", RESUME, 0);
 	r += add_key(keys, "reinstate", REINSTATE, 0);
 	r += add_key(keys, "fail", FAIL, 0);
+	r += add_key(keys, "resize", RESIZE, 0);
+	r += add_key(keys, "disablequeueing", DISABLEQ, 0);
+	r += add_key(keys, "restorequeueing", RESTOREQ, 0);
 	r += add_key(keys, "paths", PATHS, 0);
 	r += add_key(keys, "maps", MAPS, 0);
 	r += add_key(keys, "multipaths", MAPS, 0);
@@ -168,6 +172,10 @@ load_keys (void)
 	r += add_key(keys, "config", CONFIG, 0);
 	r += add_key(keys, "blacklist", BLACKLIST, 0);
 	r += add_key(keys, "devices", DEVICES, 0);
+	r += add_key(keys, "format", FMT, 1);
+	r += add_key(keys, "wildcards", WILDCARDS, 0);
+	r += add_key(keys, "quit", QUIT, 0);
+	r += add_key(keys, "exit", QUIT, 0);
 
 	if (r) {
 		free_keys(keys);
@@ -210,35 +218,40 @@ find_key (const char * str)
 static int
 get_cmdvec (char * cmd, vector *v)
 {
-	int fwd = 1;
+	int i;
 	int r = 0;
-	char * p = cmd;
+	int get_param = 0;
 	char * buff;
 	struct key * kw = NULL;
 	struct key * cmdkw = NULL;
-	vector cmdvec;
+	vector cmdvec, strvec;
+
+	strvec = alloc_strvec(cmd);
+	if (!strvec)
+		return 0;
 
 	cmdvec = vector_alloc();
 	*v = cmdvec;
 
-	if (!cmdvec)
+	if (!cmdvec) {
+		free_strvec(strvec);
 		return E_NOMEM;
+	}
 
-	while (fwd) {
-		fwd = get_word(p, &buff);
-
-		if (!buff)
-			break;
-
-		p += fwd;
+	vector_foreach_slot(strvec, buff, i) {
+		if (*buff == '"')
+			continue;
+		if (get_param) {
+			get_param = 0;
+			cmdkw->param = strdup(buff);
+			continue;
+		}
 		kw = find_key(buff);
-		FREE(buff);
-
-		if (!kw)
-			return E_SYNTAX;
-
+		if (!kw) {
+			r = E_SYNTAX;
+			goto out;
+		}
 		cmdkw = alloc_key();
-
 		if (!cmdkw) {
 			r = E_NOMEM;
 			goto out;
@@ -251,23 +264,18 @@ get_cmdvec (char * cmd, vector *v)
 		vector_set_slot(cmdvec, cmdkw);
 		cmdkw->code = kw->code;
 		cmdkw->has_param = kw->has_param;
-		
-		if (kw->has_param) {
-			if (*p == '\0')
-				goto out;
-
-			fwd = get_word(p, &buff);
-
-			if (!buff)
-				return E_NOPARM;
-
-			p += fwd;
-			cmdkw->param = buff;
-		}
+		if (kw->has_param)
+			get_param = 1;
 	}
+	if (get_param) {
+		r = E_NOPARM;
+		goto out;
+	}
+	free_strvec(strvec);
 	return 0;
 
 out:
+	free_strvec(strvec);
 	free_keys(cmdvec);
 	*v = NULL;
 	return r;
@@ -372,6 +380,7 @@ parse_cmd (char * cmd, char ** reply, int * len, void * data)
 	if (!h) {
 		*reply = genhelp_handler();
 		*len = strlen(*reply) + 1;
+		free_keys(cmdvec);
 		return 0;
 	}
 
@@ -406,15 +415,19 @@ cli_init (void) {
 		return 1;
 
 	add_handler(LIST+PATHS, NULL);
+	add_handler(LIST+PATHS+FMT, NULL);
+	add_handler(LIST+STATUS, NULL);
 	add_handler(LIST+MAPS, NULL);
 	add_handler(LIST+MAPS+STATUS, NULL);
 	add_handler(LIST+MAPS+STATS, NULL);
+	add_handler(LIST+MAPS+FMT, NULL);
 	add_handler(LIST+MAPS+TOPOLOGY, NULL);
 	add_handler(LIST+TOPOLOGY, NULL);
 	add_handler(LIST+MAP+TOPOLOGY, NULL);
 	add_handler(LIST+CONFIG, NULL);
 	add_handler(LIST+BLACKLIST, NULL);
 	add_handler(LIST+DEVICES, NULL);
+	add_handler(LIST+WILDCARDS, NULL);
 	add_handler(ADD+PATH, NULL);
 	add_handler(DEL+PATH, NULL);
 	add_handler(ADD+MAP, NULL);
@@ -423,8 +436,14 @@ cli_init (void) {
 	add_handler(RECONFIGURE, NULL);
 	add_handler(SUSPEND+MAP, NULL);
 	add_handler(RESUME+MAP, NULL);
+	add_handler(RESIZE+MAP, NULL);
+	add_handler(DISABLEQ+MAP, NULL);
+	add_handler(RESTOREQ+MAP, NULL);
+	add_handler(DISABLEQ+MAPS, NULL);
+	add_handler(RESTOREQ+MAPS, NULL);
 	add_handler(REINSTATE+PATH, NULL);
 	add_handler(FAIL+PATH, NULL);
+	add_handler(QUIT, NULL);
 
 	return 0;
 }
