@@ -24,6 +24,19 @@
 #include <vector.h>
 #include "cli.h"
 
+static void print_reply(char *s)
+{
+	if (isatty(1)) {
+		printf("%s", s);
+		return;
+	}
+	/* strip ANSI color markers */
+	while (*s != '\0') {
+		if ((*s == 0x1b) && (*(s+1) == '['))
+			while ((*s++ != 'm') && (*s != '\0')) {}; 
+		putchar(*s++);
+	}
+}
 /*
  * process the client 
  */
@@ -43,11 +56,15 @@ static void process(int fd)
 			free(line);
 			continue;
 		}
+		if (!strncmp(line, "exit", 4) && llen == 4)
+			break;
+		if (!strncmp(line, "quit", 4) && llen == 4)
+			break;
 
 		if (send_packet(fd, line, llen + 1) != 0) break;
 		if (recv_packet(fd, &reply, &len) != 0) break;
 
-		printf("%s", reply);
+		print_reply(reply);
 
 		if (line && *line)
 			add_history(line);
@@ -62,11 +79,12 @@ static void process_req(int fd, char * inbuf)
 	char *reply;
 	size_t len;
 
-	send_packet(fd, inbuf, strlen(inbuf) + 1);
-	recv_packet(fd, &reply, &len);
-
-	printf("%s", reply);
-	FREE(reply);
+	if (send_packet(fd, inbuf, strlen(inbuf) + 1) != 0)
+		return;
+	if (recv_packet(fd, &reply, &len) == 0) {
+		print_reply(reply);
+		FREE(reply);
+	}
 }
 	
 /*
